/*!
 * Collection
 *
 * Copyright (C) 2012 Philippe Plantier <philippe@midsummerweb.com>
 *
 * MIT Licensed
 */


(function(exports, global) {
    /**
     * Manages the pool of collections. This class is not to be used directly, but is used by the
     * Collection class.
     *
     * @api private
     */
    var Manager = exports.Manager = function() {
        
    };

    Manager.servers = {};
    
    /**
     * Creates a collection ID from a path and parameters.
     *
     * @api private
     */
    Manager.collectionId = function(path, params) {
        var res = encodeURIComponent(path);
        var keys = [];
        
        for (var key in params) {
            keys.push(key);
        }
        
        keys = keys.sort();
        
        for (var i = 0, l = keys.length; i < l; ++i) {
            res += "/" + encodeURIComponent(keys[i]) + "=" + encodeURIComponent(params[keys[i]]);
        }
        
        return res;
    };
    /**
     * @api private
     */
    Manager.registerCollection = function(server, path, params, collection) {
        var self = this;
        if (!(server in this.servers)) {
            this.servers[server] = {
                socket: global.io.connect(server),
                collections: {}
            };
        }
        var serverObj = this.servers[server];
        var collectionId = this.collectionId(path, params);
        
        if (!(collectionId in serverObj.collections)) {
            serverObj.socket.emit("collection", {
                operation: "subscribe",
                collection: path,
                params: params
            });
            
            serverObj.collections[collectionId] = [];
            var collectionList = serverObj.collections[collectionId];
            
            serverObj.socket.on("collection", function(msg) {
                var msgCollectionId = self.collectionId(msg.collection, msg.params);
                
                if (msgCollectionId === collectionId) {
                    for (var i in collectionList) {
                        collectionList[i].onCollection(msg);
                    }
                }
            });
        }
        
        serverObj.collections[collectionId].push(collection);
        
        return serverObj.socket;
    }
    /**
     * @api private
     */
    Manager.unregisterCollection = function(server, path, params, collection) {
        if (!(server in this.servers)) {
            return;
        }
        var serverObj = this.servers[server];
        var collectionId = this.collectionId(path, params);
        if (!(collectionId in serverObj.collections)) {
            return;
        }
        var collectionList = serverObj.collections[collectionId];
        
        for (var i = 0, l = collectionList.length; i < l; i++) {
            if (collectionList[i] === collection)
                break;
        }
        
        if (i == l)
            return;
            
        collectionList.splice(i, 1);
        if (collectionList.length === 0) {
            serverObj.socket.emit("collection", {
                operation: "unsubscribe",
                collection: path,
                params: params
            });
            
            delete serverObj.collections[collectionId];
        }
        
        // Return if there is still another registered collection
        for (i in serverObj.collections) {
            return;
        }
        
        // Do not actually disconnect the socket
        //serverObj.socket.disconnect();
        delete this.servers[server];
    }
    
    /**
     * The base class for collections.
     *
     * @param {String} server
     * @param {String} collection
     * @param {Object} options
     *
     * @api public
     */
    var Collection = exports.Collection = function(server, collection, options) {
        this.constructor(server, collection, options);
    };

    Collection.prototype._size = 10;
    Collection.prototype.increment = 10;
    Collection.prototype._starved = false;
        
    Collection.prototype.constructor = function(server, collection, options) {
        var self = this;
        
        options = options || {};
        
        this.server = server;
        this.collection = collection;
            
        options.size && (this._size = options.size);
        ("increment" in options) && (this.increment = options.increment);
        options.onChange && (this.onChange = options.onChange);
        options.onError && (this.onError = options.onError);
        options.params && (this.params = options.params);
        
        this.socket = Manager.registerCollection(server, collection, this.params, this);
        
        this._data = [];
        this._buffer = [];
        this._startOffset = 0;
        
        this._setAttach(options.attach, options.anchor, options.direction);
        this._read(this.anchor, this._size + this.increment, this.direction);
    };
        
    Collection.prototype.destroy = function() {
        Manager.unregisterCollection(this.server, this.collection, this.params, this);
    };
        
    /**
     *
     * @api private
     */
    Collection.prototype._read = function(start, count, direction) {
        if (!count)
            return;
            
        this._loading = true;
        this._expectedRead = count;
    
        var keyVal = start;
        if (keyVal === +Infinity) {
            keyVal = "+Infinity";
        }
        if (keyVal === -Infinity) {
            keyVal = "-Infinity";
        }
        
        this.socket.emit("collection", {
            collection: this.collection,
            params: this.params,
            operation: "read",
            key: keyVal,
            count: count,
            direction: this.direction
        });
    };
    
    /**
     *
     * @api private
     */
    Collection.prototype._addItem = function(position, item) {
        if (item.key < this.anchor) {
            this._startOffset ++;
        }
        this._data.splice(position, 0, item);
    };
    
    /**
     *
     * @api private
     */
    Collection.prototype._delItem = function(position) {
        if (this._data[position].key < this.anchor) {
            this._startOffset --;
        }
        this._data.splice(position, 1);
    };
    
    /**
     * Handles "read" messages.
     *
     * @param {Object} msg The message, as sent by the server
     *
     * @api private
     */
    Collection.prototype._onRead = function(msg) {
        var itemCounter = 0;
        var i, l;
        var hasInvalid = false;
        
        if (this._expectedRead > msg.data.length) {
            this._starved = true;
        }
        
        // Merge
        
        for(i = 0, l = msg.data.length; i < l; i++) {
            var item = msg.data[i];
            var i2, l2;
            var removedFrom;
            var mustAdd;
            
            if (item.invalid) {
                hasInvalid = true;
                continue;
            }
            
            for (i2 = 0, l2 = this._data.length; i2 < l2; i2++) {
                if (msg.data[i].id === this._data[i2].id)
                    break;
            }
            
            removedFrom = -1;
            
            if (i2 !== l2) {
                // We found a duplicate
                if (msg.data[i].revision > this._data[i2].revision) {
                    // Theirs is more recent. Remove ours, and push theirs to the list
                    if (i2 < itemCounter) {
                        itemCounter --;
                    }
                    removedFrom = i2;
                    mustAdd = true;
                    this._delItem(i2);
                } else {
                    mustAdd = false;
                }
            } else {
                // No dupe here, insert the item
                mustAdd = true;
            }
            
            if (mustAdd) {
                do {
                    if (itemCounter >= this._data.length)
                        break;
                        
                    if (this._data[itemCounter].key > item.key)
                        break;
                        
                    itemCounter++;
                } while (true);
                
                this._addItem(itemCounter, item);
                this._onChange(removedFrom, itemCounter, {
                    key: item.key,
                    id: item.id,
                    data: item.data
                });
                
                itemCounter++;
            }
        }
        
        if (hasInvalid) {
            this._growIfLessThan(this._size);
        } else {
            if (this._loading) {
                this._loading = false;
                
                for (i = 0, l = this._buffer.length; i < l; ++i) {
                    this.onCollection(this._buffer[i]);
                }
                this._buffer = [];
                
                if (this._attachParams) {
                    this.setAttach(this._attachParams.attach, this._attachParams.anchor, this._attachParams.direction);
                    this._attachParams = false;
                }
            }

        }
    };
    
    /**
     *
     * @api private
     */
    Collection.prototype._getInWindowCount = function() {
        if (this.direction > 0) {
            return this._data.length - this._startOffset;
        } else {
            return this._startOffset;
        }
    };
    
    /**
     *
     * @api private
     */
    Collection.prototype._growIfLessThan = function(size) {
        var i, l = this._data.length;
        
        if (this._starved)
            return false;
            
        var counter = this._getInWindowCount();

        if (counter < size) {
            var lastKey;
            if (l == 0) {
                lastKey = this.anchor;
            } else {
                lastKey = this.direction > 0 ? this._data[l-1].key : this._data[0].key;
            }
            this._read(lastKey, size - counter + this.increment, this._direction);
            return true;
        }
                    
        return false;
    };
    
    /**
     * @api private
     */
    Collection.prototype._doDelete = function(msg, refill) {
        var i, l = this._data.length;
        var deletePos;
        
        for (i = 0; i < l; ++i) {
            if (msg.id == this._data[i].id) {
                break;
            }
        }
        
        if (i == l)
            return -1;
            
        deletePos = i;
            
        this._delItem(deletePos);
        l--;
        
        if (refill) {
            this._growIfLessThan(this._size);
        }
        
        return deletePos;
    };
    
    /**
     * @api private
     */
    Collection.prototype._getBounds = function() {
        var minKey;
        var maxKey;
        var l = this._data.length;
        
        if (this.direction > 0) {
            if (l > 0) {
                minKey = this._data[0].key < this.anchor ? this._data[0].key : this.anchor;
                maxKey = this._starved ? +Infinity : this._data[l-1].key;
            } else {
                minKey = this.anchor;
                maxKey = +Infinity;
            }
        } else {
            if (l > 0) {
                maxKey = this._data[l-1].key > this.anchor ? this._data[l-1].key : this.anchor;
                minKey = this._starved ? -Infinity : this._data[0].key;
            } else {
                minKey = -Infinity;
                maxKey = this.anchor;
            }
        }
        
        return { min: minKey, max: maxKey };
    };
    
    /**
     * @api private
     */
    Collection.prototype._doInsert = function(msg, bounds) {
        var i, l = this._data.length;
        var posFound = false;
        var pos;
        
        if (!bounds) {
            bounds = this._getBounds();
        }
        
        if (msg.key < bounds.min || msg.key >= bounds.max) {
            return -1;
        }
        
        for (i = 0; i < l; ++i) {
            if (this._data[i].id === msg.id) {
                break;
            }
            
            if (!posFound && this._data[i].key > msg.key) {
                pos = i;
                posFound = true;
            }
        }
        
        
        // If the id is already on the list, ignore it.
        if (i !== l)
            return;
        
        if (!posFound) {
            pos = l;
        }
        
        this._addItem(pos, { id: msg.id, key: msg.key, revision: msg.revision, data: msg.data });
        
        return pos;
    };
    
    /**
     * @api private
     */
    Collection.prototype._onDelete = function(msg) {
        var pos = this._doDelete(msg, true);
        
        if (pos >= 0) {
            this._onChange(pos, -1, null);
        }
    };
    
    /**
     * @api private
     */
    Collection.prototype._onUpdate = function(msg) {
        var i, l = this._data.length;
        var bounds = this._getBounds();
        
        for (i = 0; i < l; ++i) {
            if (msg.id == this._data[i].id) {
                break;
            }
        }
        
        // If we know this element, and the revision we have is newer than the
        // given revision, ignore the update
        if (i !== l && this._data[i].revision >= msg.revision)
            return;
        
        // Only refill if we are moving out of current bounds
        var refill = msg.key >= bounds.max || msg.key < bounds.min;
        
        if (refill) {
            var grow = this._growIfLessThan(this._size + 1);
            
            if (grow) {
                this._buffer.push(msg);
                return;
            }
        }
        
        var deletePos = this._doDelete(msg, false);
        var insertPos = this._doInsert(msg, bounds);
        
        if (deletePos >= 0 || insertPos >= 0) {
            this._onChange(deletePos, insertPos, {
                id: msg.id,
                key: msg.key,
                data: msg.data
            });
        }
    };
    
    /**
     * Sets the attachment point of this collection.
     *
     * @param {String} attach the string "start", "end", or "key"
     * @param {Number} anchor if "attach" was set to "key", the key to attach to
     * @param {Number} direction if "attach" was set to "key", 1 to grow in the positive direction, -1 to grow in the negative direction.
     *
     * @return {Object} this for chainability
     *
     * @api private
     */
    Collection.prototype._setAttach = function(attach, anchor, direction) {
        if (this._loading) {
            this._attachParams = {
                attach: attach,
                anchor: anchor,
                direction: direction
            };
            return;
        }
        
        this.attach = attach || "start";
        
        switch (this.attach) {
        case "start":
            this.anchor = -Infinity;
            this.direction = 1;
            break;
        case "end":
            this.anchor = +Infinity;
            this.direction = -1;
            break;
        case "key":
            this.direction = direction ? direction : 1;
            this.anchor = anchor;
            break;
        default:
            throw new Error("Invalid attachment");
        }
          
        return this;
    };
        
    /**
     * @api private
     */
    Collection.prototype._onChange = function(fromIndex, toIndex, data) {
        var minBound, maxBound;

        if (fromIndex >= 0 || toIndex >= 0) {
            this.onChange(fromIndex, toIndex, data);
        }
    };
    
    /**
     * Called by the Manager when a message pertaining to this collection is issued.
     *
     * @api private
     */
    Collection.prototype.onCollection = function(msg) {
        if (this._loading && msg.operation !== "read") {
            this._buffer.push(msg);
        } else {
            switch(msg.operation) {
            case "read":
                this._onRead(msg);
                break;
            case "insert":
                this._onUpdate(msg);
                break;
            case "update":
                this._onUpdate(msg);
                break;
            case "delete":
                this._onDelete(msg);
                break;
            case "error":
                this.onError(msg.error);
                break;
            }
        }
    };
    
    /*
     * Public API
     */
    
    /**
     * Sets the new maximal visible size of the collection.
     *
     * @param size The new maximal size
     *
     * @return {Object} this for chainability
     *
     * @api public
     */
    Collection.prototype.setSize = function(size) {
        this._size = size;
        this._growIfLessThan(this._size);
        return this;
    };
    
    /**
     * Returns the size of this collection.
     *
     * @api public
     */
    Collection.prototype.getSize = function() {
        return this_size;
    };
    
    /**
     * Grows the collection by a fixed amount.
     * Shorthand method for setSize(getSize() + amount);
     *
     * @param {Number} amount The amount added to the collection size
     *
     * @return {Object} this for chainability
     *
     * @api public
     */
    Collection.prototype.grow = function(amount) {
        this.setSize(this._size + amout);
        
        return this;
    };
    
    /**
     * Insert a new element in this collection.
     *
     * @param {Number} key The key at which to insert this new element
     * @param {Object} data The data stored in this element
     */
    Collection.prototype.insert = function(key, data) {
        var self = this;
        this.socket.emit("collection", {
            collection: this.collection,
            params: self.params,
            operation: "insert",
            key: +key,
            data: data
        });
        
        return this;
    };
    
    /**
     * Insert a new element in this collection.
     *
     * @param {Number} id The id of the element to update
     * @param {Number} key The key at which to move the element
     * @param {Object} data The new data for this element
     */
    Collection.prototype.update = function(id, key, data) {
        var self = this;
        this.socket.emit("collection", {
            collection: this.collection,
            params: self.params,
            operation: "update",
            id: id,
            key: +key,
            data: data
        });
        
        return this;
    };
    
    /**
     * Insert a new element in this collection.
     *
     * @param {Number} id The id of the element to delete
     */
    Collection.prototype["delete"] = Collection.prototype.del = function(id) {
        var self = this;
        this.socket.emit("collection", {
            collection: this.collection,
            params: self.params,
            operation: "delete",
            id: id
        });
        
        return this;
    };
    
    /*
     * Public API: event notification
     */
    
    /**
     * Called when something changes in the collection. To be overridden to handle
     * such events.
     *
     * @param {Number} fromIndex The index where the item previously was,
     *                 or -1 if it was not on the collection subset
     * @param {Number} fromIndex The index where the item now is,
     *                 or -1 if it is not on the collection subset anymore.
     * @param {Object} data  The new value of the item
     */
    Collection.prototype.onChange = function(fromIndex, toIndex, data) {
    
    };
    
    /**
     * Called when an operation triggers an error.
     *
     * @param {String} error A string describing the error.
     */
    Collection.prototype.onError = function(error) {
    
    };
    

})(
    'object' === typeof module ? module.exports : (this.collection = {}),
    'object' === typeof global ? global : this
)

