/*! 
 * Collection
 *
 * Copyright (C) 2012 Philippe Plantier <philippe@midsummerweb.com>
 *
 * MIT Licensed
 */

"use strict";

/**
 * Module dependencies.
 */
 
var Static = require("socket.io").Static,
    Store = require("./store"),
    errors = require("./errors")

/**
 * Export the constructor.
 */

exports = module.exports = Manager;

/** 
 * Manager constructor.
 *
 * @param {Object} io a socket.io manager instance
 * @param {Object} options key-value store of options
 * @api public
 */
function Manager(io, options) {
    var self = this;
    
    this.routes = [];
    this.io = io;
    
    this.settings = {
        "collection path": "",
        "collection client path": __dirname + "/client"
    };
    
    for (var i in options) {
        this.settings[i] = options[i];
    }
    
    if (io.static instanceof Static) {
        io.static.add("/collection.js", {
            file: this.settings["collection client path"] + "/collection.js"
        });
    }
    
    io.of(this.settings["collection path"]).on("connection", function(socket) { self.handleConnection(socket) });

    this.io.log.info("Collection.io initialized");
}

/** 
 *
 * @api private
 */
Manager.prototype.getRoute = function(collection) {
    if (typeof collection !== "string")
        return;
    
    for (var r in this.routes) {
        var matches = this.routes[r].collection === collection;
        
        if (matches)
            break;
    }
    
    if (matches) {
        return this.routes[r];
    } 
}

/**
 * Remove invalid parameters from the parameter list.
 *
 * @api private
 */
Manager.prototype.trimParams = function(route, params) {
    var res = {};
    var key;
    
    for (key in params) {
        if (route.handler.params.indexOf(key) >= 0) {
            res[key] = params[key];
        }
    }
    
    return res;
}

/**
 * Get a unique, serialized name for the collection and its parameters.
 *
 * @api private
 */
Manager.prototype.serializeCollectionName = function(route, params) {
    var res = route.collection + "/";

    var keys = route.handler.params;

    if (keys instanceof Array && keys.length) {    
        if (typeof params !== "object") {
            return false;
        } else {
            if (!keys.every(function(item) { return item in params })) {   
                return false;
            }
            
            res += keys.map(function(key) { 
                return JSON.stringify(key) + ":" + JSON.stringify(params[key]) 
            }).join(",");
        }
    }
    
    return res;
}

/** 
 * Sends an error back to the client.
 *
 * @api private
 */
Manager.prototype.sendError = function(socket, collection, params, error) {
    // Error Types: Validation, Invalid Key, Invalid ID, Not Found, Malformed, default
    var message = { 
        collection: collection, 
        params: params,
        operation: "error",
    };
    
    switch (true) {
    case error instanceof errors.ValidationError:
        message.id = error.id;
        message.error = "validation";
        message.fields = error.fields;
        break;
        
    default:
        message.error = "internal";
        break;
    }
    
    
    socket.emit("collection", message);
}

/**
 * A wrapper agains a function, so the given function can be a synchronous function
 * that returns, a synchronous function that throws, or an asynchronous function. All
 * of those will be called in the same way, using the callback.
 *
 * @param f The function to call
 * @param args an array containing the function arguments
 * @param callback a callback to call once the function is completed.
 *
 * @api private
 */
function syncOrAsyncCall(f, args, callback) {
    var res;
    
    args.push(callback);
    try {
        res = f.apply(null, args);
    } catch(e) {
        callback(e);
        return;
    }
    
    if (res instanceof Error) {
        callback(res);
    } else {
        if (typeof res !== "undefined") {
        
            callback(null, res);
        }
    }
}

Manager.prototype.validateData = function(route, id, key, data, callback) {
    if (route.validate) {
        syncOrAsyncCall(route.validate, [id, key, data], function(err, validationFields) {
            if (err) {
                callback(err);
            } else {
                if (Object.keys(validationFields).length == 0) {
                    callback(null);
                } else {
                    callback(new errors.ValidationError(id, validationFields));
                }
            }
        });
    } else {
        callback(null, data);
    }
}

Manager.prototype.importData = function(route, data, callback) {
    if (route.import) {
        syncOrAsyncCall(route.import, [data], callback);
    } else {
        callback(null, data);
    }
}


Manager.prototype.exportData = function(route, data, callback) {
    if (route.export) {
        if (data instanceof Array) {
            var l = data.length;
            
            if (l) {
                var finished = 0;
                var error = null;
                
                data.forEach(function(record, i) {
                    syncOrAsyncCall(route.export, [record.data], function(err, exdata) {
                        finished++;
                        if (err) {
                            error = err;
                        }
                        record.data = exdata;
                        
                        if (finished == l) {
                            callback(error, data);
                        }
                    });
                });
            } else {
                syncOrAsyncCall(route.export, [], callback);
            }
        } else {
            syncOrAsyncCall(route.export, [data], callback);
        }
    } else {
        callback(null, data);
    }
}

/**
 * Validates the user message, and creates a request object from this user message.
 * This function is async, before validation and object importation can be async.
 *
 * @param route The route this message applies to
 * @param message The user message
 * @param callback A callback, to be called when the object is ready
 *
 * @api private
 */
Manager.prototype.getReq = function(route, message, callback) {
    var self = this;
    
    var req = {
        collection: message.collection,
        params: message.params,
        operation: message.operation,
    };
    
    var invalidRequest = function() {
        callback(new errors.MalformedRequestError);
    };
    
    // Validate rather strictly all parameters of the request, we only want well-formed
    // requests here
    switch (message.operation) {
    case "read":
        // Mandatory: count (number), direction (number !== 0), key (number or string)
        if (message.key === "+Infinity") {
            req.key = +Infinity;   
        } else if (message.key === "-Infinity") {
            req.key = -Infinity;
        } else {
            if (typeof message.key === "number") {
                req.key = message.key;
            } else {
                invalidRequest();
                return;
            }
        }
        
        if (typeof message.count === "number") {
            req.count = message.count;
        } else {
            invalidRequest();
            return;
        }
        
        if (typeof message.direction === "number" && message.direction !== 0) {
            req.direction = message.direction;
        } else {
            invalidRequest();
            return;
        }
        
        break;
    case "insert":
        // Mandatory: key (number), data (object)
        if (typeof message.key === "number") {
            req.key = message.key;
        } else {
            invalidRequest();
            return;
        }
        
        if (typeof message.data === "object") {
            req.data = message.data;
        } else {
            invalidRequest();
            return;
        }
        
        break;
    case "delete":
        // Mandatory: id (number or string)
        if (typeof message.id === "number" || typeof message.id === "string") {
            req.id = message.id;
        } else {
            invalidRequest();
            return;
        }
        
        break;
    case "update":
        // Mandatory: id (number or string), data (object). Optional: key (number).
        if (typeof message.id === "number" || typeof message.id === "string") {
            req.id = message.id;
        } else {
            invalidRequest();
            return;
        }
        
        if (typeof message.key === "number") {
            req.key = message.key;
        } else {
            if (typeof message.key !== "undefined") {
                invalidRequest();
                return false;
            }
        }
        
        if (typeof message.data === "object") {
            req.data = message.data;
        } else {
            invalidRequest();
            return;
        }
        
        break;
    default:
        invalidRequest();
        return;
    }
    
    if (typeof req.data !== "undefined") {
        this.validateData(route, req.id, req.key, req.data, function(err, data) {
            if (err) {
                callback(err);
            } else {
                self.importData(route, req.data, function(err, data) {
                    req.data = data;
                    callback(err, req);
                });
            }
        });
    } else {
        process.nextTick(function() {
            callback(null, req);
        });
    }
}


/** 
 *
 * @api private
 */
Manager.prototype.getRes = function(collectionId, socket, route, message, broadcast) {     
    var self = this;
       
    return function(err, descriptor) {
        if (err) {
            self.io.log.error(err);
            self.io.log.info(err.stack);
            self.sendError(socket, message.collection, message.params, err);
        } else {
            var result;
            
            if ((typeof descriptor !== "object" || descriptor === null) && message.operation !== "delete") {
                // Internal error, the storage module failed to return a valid
                // descriptor. This is a bug in the storage module.
                self.io.log.error("Internal error: invalid descriptor in send() call");
                return;
            }
            
            switch (message.operation) {
            case "read":
                result = {
                    operation: message.operation,
                    collection: message.collection,
                    params: message.params,
                    data: descriptor
                };
                break;
            case "delete":
                result = {
                    operation: message.operation,
                    collection: message.collection,
                    params: message.params,
                    id: message.id
                };
                break;
            default:
                result = {
                    operation: message.operation,
                    collection: message.collection,
                    params: message.params,
                    id: descriptor.id,
                    key: descriptor.key,
                    revision: descriptor.revision,
                    data: descriptor.data
                };
            }
            
            self.exportData(route, result.data, function(err, data) {
                if (err) {
                    self.sendError(socket, message.collection, message.params, err);
                } else {
                    result.data = data;
                    
                    if (broadcast) {
                        socket.broadcast.to("collections/" + collectionId).emit("collection", result);
                    } 
                    
                    socket.emit("collection", result); 
                }
            });
        }
    }
}

/** 
 *
 * @api private
 */
Manager.prototype.handleSubscribe = function(socket, message) {
    var route = this.getRoute(message.collection);
    var collectionId = this.serializeCollectionName(route, message.params);
    if (collectionId !== false) {
        socket.join("collections/" + collectionId);
        this.io.log.info("Joined collection " + collectionId);
    }
}

/** 
 *
 * @api private
 */
Manager.prototype.handleUnsubscribe = function(socket, message) {
    var route = this.getRoute(message.collection);
    var collectionId = this.serializeCollectionName(route, message.params);
    
    if (collectionId !== false) {
        socket.leave("collections/" + collectionId);
        this.io.log.info("Left collection " + collectionId);
    }
}

/** 
 *
 * @api private
 */
Manager.prototype.handleCollection = function(socket, message) {
    var params = {};
    var route = this.getRoute(message.collection);
    if (!route) {
        this.sendError(socket, message.collection, message.params, new errors.MalformedRequestError("Invalid Collection"));
        return;
    }
    
    var collectionId = this.serializeCollectionName(route, message.params);
    if (collectionId === false) {
        this.sendError(socket, message.collection, message.params, new errors.MalformedRequestError("Invalid collection params"));
        return;
    }
        
    // Message validation
    var op = message.operation;
    var broadcast = true;
    
    // Remove non-valid params
    message.params = this.trimParams(route, message.params);
    
    if (message.operation === "subscribe") {
        this.handleSubscribe(socket, message);
        return;
    } else if (message.operation === "unsubscribe") {
        this.handleUnsubscribe(socket, message);
        return;
    }
    
    var res = this.getRes(collectionId, socket, route, message, message.operation !== "read");
    
    this.getReq(route, message, function(err, req) {
        if (err) {
            res(err);
        } else {
            switch (op) {
            case "read":
                route.handler.read(req, res);
                break;
            case "update":
                route.handler.update(req, res);
                break;
            case "insert":
                route.handler.insert(req, res);
                break;
            case "delete":
                route.handler.delete(req, res);
                break;
            }
        }
    });
}

/** 
 *
 * @api private
 */
Manager.prototype.handleDisconnect = function(socket, message) {

}

/** 
 *
 * @api private
 */
Manager.prototype.handleConnection = function(socket) {
    var self = this;
    
    socket.on("collection", function(message) { self.handleCollection(socket, message) });
    socket.on("disconnect", function(message) { self.handleDisconnect(socket, message) });
}


/**
 * Add a new collection route.
 *
 * @api public
 */
Manager.prototype.add = function(collection, handler, options) {

    this.routes.push({
        collection: collection,
        handler: handler,
        validate: options && options.validate,
        import: options && options.import,
        export: options && options.export
    });
}
