/*! 
 * Collection
 *
 * Copyright (C) 2012 Philippe Plantier <philippe@midsummerweb.com>
 *
 * MIT Licensed
 */

"use strict";

/** 
 * Module dependencies
 */
var util = require("util"),
    mongodb = require("mongodb"),
    Store = require("../store"),
    errors = require("../errors")

/**
 * Creates a new MongoDB collection.io store.
 *
 * If options.collection a MongoDb collection, it is used as a data store.
 *
 * If options.collection is a string, and options.db is an open mongoDB Db, use said collection on the database.
 * 
 * If options.collection is a string, and options.db is a string, a new mongodb instance is open, optionally using options.host and options.port.
 *
 * Examples:
 * * `new MongoDb({host: "localhost", port: 27017, db: "dbname" collection: "myCollection"}`
 * * `new MongoDb({db: openMongoDb, collection: "myCollection"}`
 * * `new MongoDb({collection: validMongodbCollection})`
 *
 * @param {Object} options
 * 
 * @api public
 */
var MongoDb = module.exports = function(options) {
    var self = this;
    
    Store.apply(this, arguments);
    
    this._getCollectionCallbacks = [];
    
    if (typeof this.settings.collection === "string" && typeof this.settings.db === "string") {
        this.openDb();
    } else if (typeof this.settings.collection === "string" && typeof this.settings.db === "object") {
        this.db = this.settings.db;
        this.openCollection();
    } else if (typeof this.settings.collection === "object") {
        this.collection = this.settings.collection;
    } else {
        this.error = new Error("Invalid options");
    }
    
    this.getCollection(function(err, collection) {
        if (!err) {            
            self.params.forEach(function(param) {
                var index = {};
                index["params." + param] = 1;
                index.key = 1;
                collection.ensureIndex(index);
                index.key = -1;
                collection.ensureIndex(index);
            });
        }
    });
}

/**
 *
 * @api private
 */
MongoDb.prototype.openDb = function() {
    var self = this;
    
    this.db = new mongodb.Db(
        this.settings.db,
        new mongodb.Server(this.settings.host || "localhost", this.settings.port || 27017, {})
    );
    
    this.db.open(function(err, p_client) {
        if (err) {
            self.error = err;
        } else {
            self.openCollection();
        }
    });
}

/**
 *
 * @api private
 */
MongoDb.prototype.openCollection = function() {
    var self = this;

    this.db.collection(this.settings.collection, function(err, collection) {
        self.collection = collection;
        self._getCollectionCallbacks.forEach(function(cb) { cb(err, self.collection) });
    });
}

/**
 *
 * @api private
 */
MongoDb.prototype.getCollection = function(callback) {
    if (this.error) {
        callback(this.error);
    } else {
        if (!this.collection) {
            this._getCollectionCallbacks.push(callback);
        } else {
            callback(null, this.collection);
        }
    }
}

/**
 * @api private
 */
function setId(results) {
   results.forEach(function(result) {
       result.id = result._id.toString();
   }); 
}

/**
 * @api private
 */
function addParams(object, params) {
    if (params) {
        for (var key in params) {
            object["params." + key] = params[key];
        }
    }
}

/** 
 *
 */
MongoDb.prototype.read = function(req, res) {
    this.getCollection(function(err, collection) {
        if (err) {
            res(err);
        } else {
            var condition = {};
            
            if (req.key !== -Infinity && req.key !== Infinity) {
                if (req.direction > 0) {
                    condition = { key: { $gte: req.key }};
                } else {
                    condition = { key: { $lte: req.key }};
                }
            }
            addParams(condition, req.params);

            collection.find(condition)
                .sort({ key: req.direction })
                .limit(req.count)
                .toArray(function(err, results) {
                
                if (err) {
                    res(err);
                } else {
                    setId(results);
                    res(null, req.direction > 0 ? results : results.reverse());
                }
            });
        }
    });
}

/** 
 *
 */
MongoDb.prototype.insert = function(req, res) {
    this.getCollection(function(err, collection) {
        if (err) {
            res(err);
        } else {
            collection.insert({
                key: req.key,
                revision: 0,
                data: req.data,
                params: req.params
            }, function(err, docs) {
                if (err) {
                    res(err);
                } else if (!docs || !(docs.length)) {
                    res(new Error("Unable to insert"));
                } else {
                    setId(docs);
                    res(null, docs[0]);
                }
            });
        }
    });
}

/** 
 *
 */
MongoDb.prototype.delete = function(req, res) {
    this.getCollection(function(err, collection) {
        if (err) {
            res(err);
        } else {
            var id;
            
            try {
                id =  mongodb.ObjectID(req.id);
            } catch (e) {
                res(new errors.InvalidIdError(req.id));
                return;
            }
            
            var condition = {
                _id: id
            };
            
            addParams(condition, req.params);
            
            collection.findAndModify(condition, [['_id','asc']], {}, {
                remove: true
            }, function(err, result) {
                if (err) {
                    res(err);
                } else if (!result) {
                    res(new errors.NotFoundError(req.id));
                } else {
                    res(null, null);
                }
            });
        }
    });
}

/** 
 *
 */
MongoDb.prototype.update = function(req, res) {
    this.getCollection(function(err, collection) {
        if (err) {
            res(err);
        } else {
            var id;
            
            try {
                id =  mongodb.ObjectID(req.id);
            } catch (e) {
                res(new errors.InvalidIdError(req.id));
                return;
            }
            
            var condition = {
                _id: id
            };
            
            addParams(condition, req.params);
            
            collection.findAndModify(condition, [['_id','asc']], {
                $inc: { revision: 1 },
                $set: { 
                    data: req.data,
                    key: req.key 
                }
            }, {
                new: true
            }, function(err, result) {
                if (err) {
                    res(err);
                } else if (!result) {
                    res(new errors.NotFoundError(req.id));
                } else {
                    result.id = result._id.toString();
                    res(null, result);
                }
            });
        }
    });
}

/** 
 *
 */
MongoDb.prototype.clear = function(callback) {
    this.getCollection(function(err, collection) {
        collection.remove({}, callback);
    });
}
