/*! 
 * Collection
 *
 * Copyright (C) 2012 Philippe Plantier <philippe@midsummerweb.com>
 *
 * MIT Licensed
 */

"use strict";

/** 
 * Module dependencies
 */
var util = require("util"),
    redis = require("redis"),
    Store = require("../store"),
    errors = require("../errors");

function escapeToken(token) {
    token = token.toString();
    return token.replace(/[:\/=]/g, "\\$&");
}

/**
 * Creates a new Redis collection.io store.
 *
 * If options.db a Redis client, it is used as a data store.
 *
 * If options.db is a string, a new redis client is open, optionally using options.host and options.port.
 *
 * Option.key specifies the base name for the key, and is mandatory.
 *
 * Examples:
 * * `new Redis({host: "localhost", port: 6379, key: "myKey"}`
 * * `new Redis({db: redisClient, key: "myKey"}`
 *
 * @param {Object} options
 * 
 * @api public
 */
var Redis = module.exports = function(options) {
    var self = this;
    
    Store.apply(this, arguments);
    
    
    if (typeof this.settings.db === "object") {
        this.client = this.settings.db;
    } else {
        this.client = redis.createClient(this.settings.port || 6379, this.settings.host || "localhost");
    }
    
    if (typeof this.settings.key === "string") {
        this.key = this.settings.key;
    } else {
        this.error = new Error("Missing mandatory option.key");
    }
};

/**
 *
 * @api private
 */
Redis.prototype.getBaseKeyName = function(req) {
    var i, l = this.params ? this.params.length : 0;
    var res = escapeToken(this.key) + ":";
    
    for (i = 0; i < l; ++i) {
        res += escapeToken(this.params[i]);
        res += "=";
        res += escapeToken(req.params[this.params[i]]);
        res += ":";
    }
    
    return res;
};

/** 
 *
 */
Redis.prototype.read = function(req, res) {
    var command;
    var bound;
    var self = this;
    
    if (req.direction > 0) {
        command = "zrangebyscore";
        bound = "Inf";
    } else {
        command = "zrevrangebyscore";
        bound = "-Inf";
    }
    this.client[command](this.getBaseKeyName(req) + "sort", req.key, bound, "WITHSCORES", "LIMIT", 0, req.count, function(err, reply) {
        if (err) {
            res(err);
        } else {
            var i, l = reply.length;
            var descriptors = [];
            
            for (i = 0; i < l; i += 2) {                
                descriptors.push({
                    key: +(reply[i+1]),
                    id: reply[i]
                });   
            }
            
            var counter = descriptors.length;
            var readError = false;
            
            var readFinished = function(err, obj, descriptor) {
                --counter;
                
                if (err) {
                    readError = err;
                } else {
                    if (obj && +obj.key === descriptor.key) {
                        descriptor.revision = obj.revision;
                        
                        try {
                            descriptor.data = JSON.parse(obj.data);
                        } catch (err) {
                            descriptor.data = {};
                        }
                    } else {
                        descriptor.invalid = true;
                    }
                }
                
                if (!counter) {
                    if (readError) {
                        res(readError);
                    } else {
                        if (req.direction < 0) {
                            res(null, descriptors.reverse());
                        } else {
                            res(null, descriptors);
                        }
                    }
                }
            };
            
            if (descriptors.length) {
                descriptors.forEach(function(descriptor) {
                    self.client.hgetall(self.getBaseKeyName(req) + "id:" + escapeToken(descriptor.id), function(err, obj) {
                        readFinished(err, obj, descriptor);
                    });
                });
            } else {
                res(null, descriptors);
            }
        }
    });
};

/** 
 *
 */
Redis.prototype.insert = function(req, res) {
    var self = this;
    
    this.client.incr(this.getBaseKeyName(req) + "counter", function(err, id) {
        if (err) {
            res(err);
        } else {
            self.client.multi()
                .hmset(self.getBaseKeyName(req) + "id:" + escapeToken(id), {
                    revision: 0,
                    data: JSON.stringify(req.data),
                    key: req.key
                })
                .zadd(self.getBaseKeyName(req) + "sort", req.key, id)
                .exec(function(err) {
                    if (err) {
                        res(err);
                    } else {
                        res(null, {
                            id: id,
                            revision: 0,
                            key: req.key,
                            data: req.data
                        });
                    }
                })
            ;   
        }
    });
};

/** 
 *
 */
Redis.prototype.doDelete = function(baseKey, sortKey, id, callback) {
    var self = this;
    
    this.client.multi()
        .del(baseKey)
        .zrem(sortKey, id)
        .exec(function(err, replies) {
            if (err) {
                callback(err);
            } else {
                if (replies[1] == 0) {
                    callback(new errors.NotFoundError(id));
                } else {
                    callback(null);
                }
            }
        })
    ;
};

/** 
 *
 */
Redis.prototype.delete = function(req, res) {
    var self = this;
    var baseKey = this.getBaseKeyName(req) + "id:" + escapeToken(req.id);
    var sortKey = this.getBaseKeyName(req) + "sort";

    this.doDelete(baseKey, sortKey, req.id, function(err) {
        if (err) {
            res(err);
        } else {
            res(null, null);
        }
    });
};

/** 
 *
 */
Redis.prototype.update = function(req, res) {
    var self = this;
    
    var baseKey = this.getBaseKeyName(req) + "id:" + escapeToken(req.id);
    var sortKey = this.getBaseKeyName(req) + "sort";

    self.client.multi()
        .hexists(baseKey, "key")
        .zrem(sortKey, req.id)
        .zadd(sortKey, req.key, req.id)
        .hmset(baseKey, {
            data: JSON.stringify(req.data),
            key: req.key
        })
        .hincrby(baseKey, "revision", 1)
        .exec(function(err, replies) {
            if (err) {
                res(err);
                return;
            }
            
            if (replies[0] === 0) {
                self.doDelete(baseKey, sortKey, req.id, function (err) {
                    if (err) {
                        res(err);
                    } else {
                        res(new errors.NotFoundError(req.id));
                    }
                }); 
            } else {
                res(null, {
                    id: req.id,
                    key: req.key,
                    data: req.data,
                    revision: replies[4] 
                });
            }
            
        })
    ;
};

/** 
 *
 */
Redis.prototype.clear = function(callback) {
    var self = this;
    
    this.client.keys(escapeToken(this.key) + ":*", function(err, results) {
        var keycount = results.length;
        
        results.forEach(function (result) {
            self.client.del(result, function(err, res) {
                keycount--;
                
                if (!keycount) {
                    callback();
                }
            });
        });
    
        if (!results || !results.length) {
            callback();
        }
    });
}
