/*! 
 * Collection
 *
 * Copyright (C) 2012 Philippe Plantier <philippe@midsummerweb.com>
 *
 * MIT Licensed
 */

var basedir;

if (process.env.JSCOVERAGE) {
    basedir = "../../var/jscoverage";
} else {
    basedir = "../../lib";
}

var vows = require("vows"),
    mock = require("./lib/mock"),
    assert = require("assert"),
    async = require("async"),
    client = require(basedir + "/client/collection");

function onChangeHandler(a) {
    return function(from, to, data) {
        if (!a.history) {
            a.history = [];
        }
        a.history.push({from: from, to: to, data: data ? data.data : null});
        
        if (from >= 0) {
            a.splice(from, 1);
        }
        if (to >= 0) {
            a.splice(to, 0, data.data);
        }
    }
}
  
vows.describe("Client")
.addBatch({
    "Creation": {
        topic: function() {
            var res = [];
            var self = this;
            
            global.io = new mock.Io(function(err, message) {
                if (err) {
                    this.callback(err);
                } else {
                    res.push(message);
                    if (res.length == 2) {
                        self.callback(null, res);
                    }
                }
            });
            new client.Collection("server", "path");
        },
        "Subscribe was called": function(err, message) {
            assert.equal(message[0].data.operation, "subscribe");
            assert.equal(message[0].data.collection, "path");
        },
        "Read was called": function(err, message) {
            assert.equal(message[1].data.operation, "read");
            assert.equal(message[1].data.collection, "path");
        }
    }
})
.addBatch({
    "": {
        topic: function() {
            global.io = new mock.Io();
        }
    }
})
.addBatch({
    "_onRead() - basic": {
        topic: function() {
            var self = this;
            
            var res = new client.Collection("server", "path", {
                size: 1,
                increment: 1
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } },
                    { id: 1,  key: 1, revision: 0, data: { t: "data1" } },
                ]
            })
            return res;
        },
        "Is not starved": function(err, coll) {
            assert.equal(coll._starved, false);  
        },
        "Has 2 elements": function(err, coll) {
            assert.equal(coll._data.length, 2);
        }
    },
    "_onRead() - starved": {
        topic: function() {
            var self = this;

            var res = new client.Collection("server", "path", {
                size: 1,
                increment: 1
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } }
                ]
            })
            return res;
        },
        "Is starved": function(err, coll) {
            assert.equal(coll._starved, true);  
        },
        "Has 1 element": function(err, coll) {
            assert.equal(coll._data.length, 1);
        }
    },   
    "insert + read combo, insert is more recent": {
        topic: function() {
            var self = this;
            var a = [];
            
            var res = new client.Collection("server", "path", {
                size: 2,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "insert",
                collection: "path",
                id: 1,
                key: 1,
                revision: 1,
                data: { t: "data1.1" }
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } },
                    { id: 1,  key: 1, revision: 0, data: { t: "data1" } },
                    { id: 2,  key: 2, revision: 0, data: { t: "data2" } }
                ]
            });
            return { a: a, coll: res };
        },
        "Is not starved": function(err, res) {
            assert.equal(res.coll._starved, false);  
        },
        "Has 3 elements": function(err, res) {
            assert.equal(res.coll._data.length, 3);
        },
        "Second element is the newer one": function(err, res) {
            assert.equal(res.coll._data[1].revision, 1);
            assert.equal(res.coll._data[1].data.t, "data1.1");
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data1.1");
            assert.equal(res.a[2].t, "data2");
        }
    },
    "insert + read combo, read is more recent": {
        topic: function() {
            var self = this;
            var a = [];

            var res = new client.Collection("server", "path", {
                size: 2,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "insert",
                collection: "path",
                id: 1,
                key: 1,
                revision: 0,
                data: { t: "data1" }
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } },
                    { id: 1,  key: 1, revision: 1, data: { t: "data1.1" } },
                    { id: 2,  key: 2, revision: 0, data: { t: "data2" } }
                ]
            });
            
            return { a: a, coll: res };
        },
        "Is not starved": function(err, res) {
            assert.equal(res.coll._starved, false);  
        },
        "Has 3 elements": function(err, res) {
            assert.equal(res.coll._data.length, 3);
        },
        "Second element is the newer one": function(err, res) {
            assert.equal(res.coll._data[1].revision, 1);
            assert.equal(res.coll._data[1].data.t, "data1.1");
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data1.1");
            assert.equal(res.a[2].t, "data2");
        }
    },
    "insert inside": {
        topic: function() {
            var self = this;
            var a = [];
            
            var res = new client.Collection("server", "path", {
                size: 1,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } },
                    { id: 2,  key: 2, revision: 0, data: { t: "data2" } }
                ]
            });
            
            res.onCollection({
                operation: "insert",
                collection: "path",
                id: 1,
                key: 1,
                revision: 0,
                data: { t: "data1" }
            });
            
            return { a: a, coll: res };
        },

        "Has 3 elements": function(err, res) {
            assert.equal(res.coll._data.length, 3);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data1");
            assert.equal(res.a[2].t, "data2");
        }
    },
    "insert after (not starved)": {
        topic: function() {
            var self = this;
            var a = [];
            
            var res = new client.Collection("server", "path", {
                size: 1,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } },
                    { id: 2,  key: 2, revision: 0, data: { t: "data2" } }
                ]
            });
            
            res.onCollection({
                operation: "insert",
                collection: "path",
                id: 1,
                key: 3,
                revision: 0,
                data: { t: "data1" }
            });
            
            return { a: a, coll: res };
        },

        "Has 2 elements": function(err, res) {
            assert.equal(res.coll._data.length, 2);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data2");
        }
    },
    "insert after (starved)": {
        topic: function() {
            var self = this;
            var a = [];
            
            var res = new client.Collection("server", "path", {
                size: 2,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } },
                    { id: 2,  key: 2, revision: 0, data: { t: "data2" } }
                ]
            });
            
            res.onCollection({
                operation: "insert",
                collection: "path",
                id: 1,
                key: 3,
                revision: 0,
                data: { t: "data1" }
            });
            
            return { a: a, coll: res };
        },

        "Has 3 elements": function(err, res) {
            assert.equal(res.coll._data.length, 3);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data2");
            assert.equal(res.a[2].t, "data1");
        }
    },
    "insert before (not starved, anchored to start)": {
        topic: function() {
            var self = this;
            var a = [];
            
            var res = new client.Collection("server", "path", {
                size: 2,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } },
                    { id: 2,  key: 2, revision: 0, data: { t: "data2" } }
                ]
            });
            
            res.onCollection({
                operation: "insert",
                collection: "path",
                id: 1,
                key: -1,
                revision: 0,
                data: { t: "data1" }
            });
            return { a: a, coll: res };
        },

        "Has 3 elements": function(err, res) {
            assert.equal(res.coll._data.length, 3);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data1");
            assert.equal(res.a[1].t, "data0");
            assert.equal(res.a[2].t, "data2");
        }
    },
    "insert before (starved, anchored to key 0)": {
        topic: function() {
            var self = this;
            var a = [];
            
            var res = new client.Collection("server", "path", {
                attach: "key",
                anchor: 0,
                size: 2,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } },
                    { id: 2,  key: 2, revision: 0, data: { t: "data2" } }
                ]
            });
            
            res.onCollection({
                operation: "insert",
                collection: "path",
                id: 1,
                key: -1,
                revision: 0,
                data: { t: "data1" }
            });
            return { a: a, coll: res };
        },

        "Has 2 elements": function(err, res) {
            assert.equal(res.coll._data.length, 2);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data2");
        }
    },
    "insert before (starved, anchored to key 0, inverted)": {
        topic: function() {
            var self = this;
            var a = [];
            
            var res = new client.Collection("server", "path", {
                attach: "key",
                anchor: 0,
                direction: -1,
                size: 2,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: -2, revision: 0, data: { t: "data0" } },
                    { id: 2,  key: -1, revision: 0, data: { t: "data2" } }
                ]
            });
            
            res.onCollection({
                operation: "insert",
                collection: "path",
                id: 1,
                key: -3,
                revision: 0,
                data: { t: "data1" }
            });
            return { a: a, coll: res };
        },

        "Has 3 elements": function(err, res) {
            assert.equal(res.coll._data.length, 3);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data1");
            assert.equal(res.a[1].t, "data0");
            assert.equal(res.a[2].t, "data2");
        }
    },
    "insert after (starved, anchored to key 0, inverted)": {
        topic: function() {
            var self = this;
            var a = [];
            
            var res = new client.Collection("server", "path", {
                attach: "key",
                anchor: 0,
                direction: -1,
                size: 2,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: -2, revision: 0, data: { t: "data0" } },
                    { id: 2,  key: -1, revision: 0, data: { t: "data2" } }
                ]
            });
            
            res.onCollection({
                operation: "insert",
                collection: "path",
                id: 1,
                key: 2,
                revision: 0,
                data: { t: "data1" }
            });
            return { a: a, coll: res };
        },

        "Has 2 elements": function(err, res) {
            assert.equal(res.coll._data.length, 2);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data2");
        }
    },
    "update inside": {
        topic: function() {
            var self = this;
            var a = [];
            
            var res = new client.Collection("server", "path", {
                size: 3,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } },
                    { id: 1,  key: 1, revision: 0, data: { t: "data1" } },
                    { id: 2,  key: 2, revision: 0, data: { t: "data2" } },
                    { id: 3,  key: 3, revision: 0, data: { t: "data3" } },
                ]
            });
            
            res.onCollection({
                operation: "update",
                collection: "path",
                id: 1,
                key: 2.2,
                revision: 1,
                data: { t: "data1" }
            });
            
            return { a: a, coll: res };
        },

        "Has 4 elements": function(err, res) {
            assert.equal(res.coll._data.length, 4);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data2");
            assert.equal(res.a[2].t, "data1");
            assert.equal(res.a[3].t, "data3");
        }
    },
    "update before": {
        topic: function() {
            var self = this;
            var a = [];
            
            var res = new client.Collection("server", "path", {
                size: 3,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } },
                    { id: 1,  key: 1, revision: 0, data: { t: "data1" } },
                    { id: 2,  key: 2, revision: 0, data: { t: "data2" } },
                    { id: 3,  key: 3, revision: 0, data: { t: "data3" } },
                ]
            });
            
            res.onCollection({
                operation: "update",
                collection: "path",
                id: 1,
                key: -1,
                revision: 1,
                data: { t: "data1" }
            });
            
            return { a: a, coll: res };
        },

        "Has 4 elements": function(err, res) {
            assert.equal(res.coll._data.length, 4);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data1");
            assert.equal(res.a[1].t, "data0");
            assert.equal(res.a[2].t, "data2");
            assert.equal(res.a[3].t, "data3");
        }
    },
    "update after, not starvable": {
        topic: function() {
            var self = this;
            var a = [];
            
            var res = new client.Collection("server", "path", {
                size: 3,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } },
                    { id: 1,  key: 1, revision: 0, data: { t: "data1" } },
                    { id: 2,  key: 2, revision: 0, data: { t: "data2" } },
                    { id: 3,  key: 3, revision: 0, data: { t: "data3" } },
                ]
            });
            
            res.onCollection({
                operation: "update",
                collection: "path",
                id: 1,
                key: 4,
                revision: 1,
                data: { t: "data1" }
            });
            
            return { a: a, coll: res };
        },

        "Has 3 elements": function(err, res) {
            assert.equal(res.coll._data.length, 3);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data2");
            assert.equal(res.a[2].t, "data3");
        }
    },
    "update after, starvable": {
        topic: function() {
            var self = this;
            var a = [];
            
            var res = new client.Collection("server", "path", {
                size: 3,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } },
                    { id: 1,  key: 1, revision: 0, data: { t: "data1" } },
                    { id: 2,  key: 2, revision: 0, data: { t: "data2" } },
                    { id: 3,  key: 3, revision: 0, data: { t: "data3" } },
                ]
            });
            
            res.onCollection({
                operation: "delete",
                collection: "path",
                id: 2
            });
            
            res.onCollection({
                operation: "update",
                collection: "path",
                id: 1,
                key: 4.5,
                revision: 1,
                data: { t: "data1" }
            });
            
            if (res._loading) {
                res.onCollection({
                    operation: "read",
                    collection: "path",
                    data: [
                        { id: 4,  key: 4, revision: 0, data: { t: "data4" } },
                        { id: 5,  key: 5, revision: 0, data: { t: "data5" } }
                    ]
                }); 
            } else {
                return new Error("Should be loading now");
            }
            
            return { a: a, coll: res };
        },

        "Is not errored": function(err, res) {
            assert(!(res instanceof Error));
        },
        "Has 5 elements": function(err, res) {
            assert.equal(res.coll._data.length, 5);
        },
        "Synced array is correct": function(err, res) {            
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data3");
            assert.equal(res.a[2].t, "data4");
            assert.equal(res.a[3].t, "data1");
            assert.equal(res.a[4].t, "data5");
        }
    },
    "multiple reads with duplicates": {
        topic: function() {
            var self = this;
            var a = [];
            
            var res = new client.Collection("server", "path", {
                size: 3,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } },
                    { id: 1,  key: 1, revision: 0, data: { t: "data1" } },
                    { id: 2,  key: 2, revision: 0, data: { t: "data2" } },
                    { id: 3,  key: 3, revision: 0, data: { t: "data3" } },
                ]
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 3,  key: 1.5, revision: 1, data: { t: "data3" } },
                    { id: 4,  key: 4, revision: 0, data: { t: "data4" } }
                ]
            }); 
            
            
            return { a: a, coll: res };
        },

        "Has 4 elements": function(err, res) {
            assert.equal(res.coll._data.length, 5);
        },
        "Synced array is correct": function(err, res) {        
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data1");
            assert.equal(res.a[2].t, "data3");
            assert.equal(res.a[3].t, "data2");
            assert.equal(res.a[4].t, "data4");
        },
        "Callback history is correct": function(err, res) {        
            assert.deepEqual(res.a.history[3], { from: -1, to: 3, data: { t: 'data3' }});
            assert.deepEqual(res.a.history[4], { from: 3, to: 2, data: { t: 'data3' }});
        },
    },
    "multiple reads with duplicates - 2": {
        topic: function() {
            var self = this;
            var a = [];
            
            var res = new client.Collection("server", "path", {
                size: 3,
                increment: 1,
                onChange: onChangeHandler(a)
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0,  key: 0, revision: 0, data: { t: "data0" } },
                    { id: 1,  key: 1, revision: 0, data: { t: "data1" } },
                    { id: 2,  key: 2, revision: 0, data: { t: "data2" } },
                    { id: 3,  key: 3, revision: 0, data: { t: "data3" } },
                ]
            });
            
            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 4,  key: 4, revision: 0, data: { t: "data4" } },
                    { id: 2,  key: 4.5, revision: 1, data: { t: "data2.1" } },
                    { id: 5,  key: 5, revision: 0, data: { t: "data5" } }
                ]
            }); 
            
            
            return { a: a, coll: res };
        },

        "Has 4 elements": function(err, res) {
            assert.equal(res.coll._data.length, 6);
        },
        "Synced array is correct": function(err, res) {        
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data1");
            assert.equal(res.a[2].t, "data3");
            assert.equal(res.a[3].t, "data4");
            assert.equal(res.a[4].t, "data2.1");
            assert.equal(res.a[5].t, "data5");
        },
        "Callback history is correct": function(err, res) {        
            assert.deepEqual(res.a.history[5], { from: 2, to: 4, data: { t: 'data2.1' }});
        },
    },
    "partially invalid read": {
        topic: function() {
            var self = this;
            var a = [];
            var res;
            
            res = new client.Collection("server", "path", {
                size: 3,
                increment: 0,
                onChange: onChangeHandler(a)
            });
    
            res.socket = new mock.Socket(function (err, msg) {
                if (msg.message === "collection" && msg.data.operation === "read") {
                    res.onCollection({
                        operation: "read",
                        collection: "path",
                        data: [
                            { id: 2, key: 2, revision: 0, data: { t: "data2" } },
                        ]
                    });
                }
            });

            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0, key: 0, revision: 0, data: { t: "data0" } },
                    { id: 1, key: 1, revision: 0, data: { t: "data1" } },
                    { invalid: true }
                ]
            });
            
            return { a: a, coll: res };
        },

        "Has 3 elements": function(err, res) {
            assert.equal(res.coll._data.length, 3);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data1");
            assert.equal(res.a[2].t, "data2");
        }
    },
    "partially invalid read because of update at the last position": {
        topic: function() {
            var self = this;
            var a = [];
            var res;
            
            res = new client.Collection("server", "path", {
                size: 3,
                increment: 0,
                onChange: onChangeHandler(a)
            });
    
            res.socket = new mock.Socket(function (err, msg) {
                if (msg.message === "collection" && msg.data.operation === "read") {
                    res.onCollection({
                        operation: "read",
                        collection: "path",
                        data: [
                            { id: 2, key: 4, revision: 1, data: { t: "data2" } },
                        ]
                    });
                }
            });

            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0, key: 0, revision: 0, data: { t: "data0" } },
                    { id: 1, key: 1, revision: 0, data: { t: "data1" } },
                    { invalid: true }
                ]
            });
            
            res.onCollection({
                operation: "update",
                collection: "path",
                id: 2,
                key: 4,
                revision: 1,
                data: { t: "data2" }
            });
            
            return { a: a, coll: res };
        },

        "Has 3 elements": function(err, res) {
            assert.equal(res.coll._data.length, 3);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data1");
            assert.equal(res.a[2].t, "data2");
        }
    },
    "partially invalid read because of update from the last position to an inside position": {
        topic: function() {
            var self = this;
            var a = [];
            var res;
            
            res = new client.Collection("server", "path", {
                size: 3,
                increment: 0,
                onChange: onChangeHandler(a)
            });
    
            res.socket = new mock.Socket(function (err, msg) {
                if (msg.message === "collection" && msg.data.operation === "read") {
                    res.onCollection({
                        operation: "read",
                        collection: "path",
                        data: [
                            { id: 3, key: 3, revision: 0, data: { t: "data3" } },
                        ]
                    });
                }
            });

            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0, key: 0, revision: 0, data: { t: "data0" } },
                    { id: 1, key: 1, revision: 0, data: { t: "data1" } },
                    { invalid: true }
                ]
            });
            
            res.onCollection({
                operation: "update",
                collection: "path",
                id: 2,
                key: 0.5,
                revision: 1,
                data: { t: "data2" }
            });
            
            return { a: a, coll: res };
        },

        "Has at least 3 elements": function(err, res) {
            assert(res.coll._data.length >= 3);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data2");
            assert.equal(res.a[2].t, "data1");
        }
    },
    "partially invalid read because of an inside update": {
        topic: function() {
            var self = this;
            var a = [];
            var res;
            
            res = new client.Collection("server", "path", {
                size: 3,
                increment: 0,
                onChange: onChangeHandler(a)
            });
    
            res.socket = new mock.Socket(function (err, msg) {
                if (msg.message === "collection" && msg.data.operation === "read") {
                    res.onCollection({
                        operation: "read",
                        collection: "path",
                        data: [
                            { id: 3, key: 3, revision: 0, data: { t: "data3" } },
                        ]
                    });
                }
            });

            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0, key: 0, revision: 0, data: { t: "data0" } },
                    { invalid: true },
                    { id: 2, key: 2, revision: 0, data: { t: "data2" } }
                ]
            });
            
            res.onCollection({
                operation: "update",
                collection: "path",
                id: 1,
                key: 1,
                revision: 1,
                data: { t: "data1" }
            });
            
            return { a: a, coll: res };
        },

        "Has at least 3 elements": function(err, res) {
            assert(res.coll._data.length >= 3);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data1");
            assert.equal(res.a[2].t, "data2");
        }
    },
    "partially invalid read because of an update from the inside to the last pos": {
        topic: function() {
            var self = this;
            var a = [];
            var res;
            
            res = new client.Collection("server", "path", {
                size: 3,
                increment: 0,
                onChange: onChangeHandler(a)
            });
    
            res.socket = new mock.Socket(function (err, msg) {
                if (msg.message === "collection" && msg.data.operation === "read") {
                    res.onCollection({
                        operation: "read",
                        collection: "path",
                        data: [
                            { id: 1, key: 3, revision: 1, data: { t: "data1" } },
                        ]
                    });
                }
            });

            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0, key: 0, revision: 0, data: { t: "data0" } },
                    { invalid: true },
                    { id: 2, key: 2, revision: 0, data: { t: "data2" } }
                ]
            });
            
            res.onCollection({
                operation: "update",
                collection: "path",
                id: 1,
                key: 3,
                revision: 1,
                data: { t: "data1" }
            });
            
            return { a: a, coll: res };
        },

        "Has at least 3 elements": function(err, res) {
            assert(res.coll._data.length >= 3);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data2");
            assert.equal(res.a[2].t, "data1");
        }
    },
    "partially invalid read because of an update from the inside to after the last pos": {
        topic: function() {
            var self = this;
            var a = [];
            var res;
            
            res = new client.Collection("server", "path", {
                size: 3,
                increment: 0,
                onChange: onChangeHandler(a)
            });
    
            res.socket = new mock.Socket(function (err, msg) {
                if (msg.message === "collection" && msg.data.operation === "read") {
                    res.onCollection({
                        operation: "read",
                        collection: "path",
                        data: [
                            { id: 3, key: 3, revision: 1, data: { t: "data3" } },
                        ]
                    });
                }
            });

            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0, key: 0, revision: 0, data: { t: "data0" } },
                    { invalid: true },
                    { id: 2, key: 2, revision: 0, data: { t: "data2" } }
                ]
            });
            
            res.onCollection({
                operation: "update",
                collection: "path",
                id: 1,
                key: 9999,
                revision: 1,
                data: { t: "data1" }
            });
            
            return { a: a, coll: res };
        },

        "Has at least 3 elements": function(err, res) {
            assert.equal(res.coll._data.length, 3);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data2");
            assert.equal(res.a[2].t, "data3");
        }
    },
    "completely invalid read": {
        topic: function() {
            var self = this;
            var a = [];
            var res;
            
            res = new client.Collection("server", "path", {
                size: 2,
                increment: 0,
                onChange: onChangeHandler(a)
            });
    
            res.socket = new mock.Socket(function (err, msg) {
                if (msg.message === "collection" && msg.data.operation === "read") {
                    res.onCollection({
                        operation: "read",
                        collection: "path",
                        data: [
                            { id: 0, key: 2, revision: 1, data: { t: "data0" } },
                            { id: 1, key: 3, revision: 1, data: { t: "data1" } }
                        ]
                    });
                }
            });

            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { invalid: true },
                    { invalid: true }
                ]
            });
            
            res.onCollection({
                operation: "update",
                collection: "path",
                id: 0,
                key: 2,
                revision: 1,
                data: { t: "data0" }
            });
            
            res.onCollection({
                operation: "update",
                collection: "path",
                id: 1,
                key: 3,
                revision: 1,
                data: { t: "data1" }
            });
            
            return { a: a, coll: res };
        },

        "Has 2 elements": function(err, res) {
            assert.equal(res.coll._data.length, 2);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data1");
        }
    },
    "setSize": {
        topic: function() {
            var self = this;
            var a = [];
            var res;
            
            res = new client.Collection("server", "path", {
                size: 2,
                increment: 0,
                onChange: onChangeHandler(a)
            });
    
            res.socket = new mock.Socket(function (err, msg) {
                if (msg.message === "collection" && msg.data.operation === "read") {
                    res.onCollection({
                        operation: "read",
                        collection: "path",
                        data: [
                            { id: 2, key: 2, revision: 0, data: { t: "data2" } },
                            { id: 3, key: 3, revision: 0, data: { t: "data3" } }
                        ]
                    });
                }
            });

            res.onCollection({
                operation: "read",
                collection: "path",
                data: [
                    { id: 0, key: 0, revision: 0, data: { t: "data0" } },
                    { id: 1, key: 1, revision: 0, data: { t: "data1" } }
                ]
            });
            
            res.setSize(4);
            
            return { a: a, coll: res };
        },

        "Has 4 elements": function(err, res) {
            assert.equal(res.coll._data.length, 4);
        },
        "Synced array is correct": function(err, res) {
            assert.equal(res.a[0].t, "data0");
            assert.equal(res.a[1].t, "data1");
            assert.equal(res.a[2].t, "data2");
            assert.equal(res.a[3].t, "data3");
            
        }
    },
})
.export(module);
