/*! 
 * Collection
 *
 * Copyright (C) 2012 Philippe Plantier <philippe@midsummerweb.com>
 *
 * MIT Licensed
 */

var basedir;

if (process.env.JSCOVERAGE) {
    basedir = "../../var/jscoverage";
} else {
    basedir = "../../lib";
}

var vows = require("vows"),
    assert = require("assert"),
    util = require("util"),
    collection = require(basedir + "/collection"),
    Manager = collection.Manager,
    Store = collection.Store,
    mock = require("./lib/mock");

function getManager(callback) {
    var manager = new Manager(new mock.Io);
    manager.add("Test1");
    manager.add("Test2", new mock.Store());
    manager.add("Test3", new mock.Store({ params: ["foo", "bar"] }, callback));
    manager.add("Test4", new mock.Store({ params: ["bar", "foo"] }, callback));
    
    return manager;
}

vows.describe("Manager")
.addBatch({
    "add()": {
        topic: function() {
            var manager = getManager();
            return manager;
        },
        
        "Has 3 routes": function(err, manager) {
            assert.equal(manager.routes.length, 4);
        },
        
        "Routes have the correct order": function(err, manager) {
            assert.equal(manager.routes[0].collection, "Test1");
            assert.equal(manager.routes[1].collection, "Test2");
            assert.equal(manager.routes[2].collection, "Test3");
            assert.equal(manager.routes[3].collection, "Test4");
        },
        
        "Second route has the correct handler": function(err, manager) {
            assert.ok(manager.routes[1].handler instanceof mock.Store);
        }
    },
    
    "Routes": {
        topic: function() {
            return getManager();
        },
        
        "getRoute()": function(err, manager) {
            var route = manager.getRoute("Test2");
            assert.equal(route.collection, "Test2");
            assert.equal(typeof route.handler, "object");
        },
        
        "serializeCollectionName() - 1": function(err, manager) {
            var route = manager.getRoute("Test3");
            var name = manager.serializeCollectionName(route, {bar: 1, foo: 1});
            assert.equal(name, 'Test3/"foo":1,"bar":1');
        },
        
        "serializeCollectionName() - order": function(err, manager) {
            var route = manager.getRoute("Test4");
            var name = manager.serializeCollectionName(route, {bar: 1, foo: 1});
            assert.equal(name, 'Test4/"bar":1,"foo":1');
        },
        
        "serializeCollectionName() - missing parameter": function(err, manager) {
            var route = manager.getRoute("Test4");
            var name = manager.serializeCollectionName(route, {bar: 1});
            assert.equal(name, false);
        },
        
        "serializeCollectionName() - invalid parameters": function(err, manager) {
            var route = manager.getRoute("Test4");
            var name = manager.serializeCollectionName(route, 0);
            assert.equal(name, false);
        },
        
    },
    
    "getRes(), read, no broadcast": {
        topic: function() {
            var manager = getManager();
            var socket = new mock.Socket();
            var res = manager.getRes("cid", socket, {}, {
                operation: "read",
                collection: "coll",
                params: {}
            }, false);
            res(null, []);
            return socket;
            
        },
        
        "One single emit": function(err, socket) {
            assert.equal(socket.operations.length, 1);
        },
        
        "Was not broadcast": function(err, socket) {
            assert.notEqual(socket.operations[0].broadcast, true);
        },
        
        "Has the correct data": function(err, socket) {
            assert.ok(socket.operations[0].data.data instanceof Array);
            assert.equal(socket.operations[0].data.data.length, 0);
        }
    },
    
    "getRes(), insert, broadcast": {
        topic: function() {
            var manager = getManager();
            var socket = new mock.Socket();
            var res = manager.getRes("cid", socket, {}, {
                operation: "insert",
                collection: "coll",
                params: {}
            }, true);
            res(null, {
                id: "id",
                key: "key",
                revision: 0,
                data: "data"
            });
            return socket;
            
        },
        
        "At least one emit": function(err, socket) {
            assert.ok(socket.operations.length >= 1);
        },
        
        "An emit was broadcast": function(err, socket) {
            var broadcast = socket.operations.some(function(item) { return item.broadcast });
            
            assert.ok(broadcast);
        },
        
        "Has the correct data": function(err, socket) {
            var ok = socket.operations.every(function(item) {
                return item.data.id === "id" &&
                     item.data.key === "key" &&
                     item.data.revision === 0 &&
                     item.data.data === "data";
            });
            assert.ok(ok);
        }
    },
    
    "handleCollection(), read from Infinity": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "read",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                key: "-Infinity",
                count: 10,
                direction: 1
            });
        },
        
        "Read operation was called": function(err, operation) {
            assert.equal(operation.req.operation, "read");
        },
        
        "Key is -Infinity": function(err, operation) {
            assert.strictEqual(operation.req.key, -Infinity);
        },
        
        "Direction is 1": function(err, operation) {
            assert.strictEqual(operation.req.direction, 1);
        },
        
        "Count is 10": function(err, operation) {
            assert.strictEqual(operation.req.count, 10);
        }
    },
    "handleCollection(), read from number": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "read",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                key: 10,
                count: 10,
                direction: 1
            });
        },
        
        "Read operation was called": function(err, operation) {
            assert.equal(operation.req.operation, "read");
        },
        
        "Key is 10": function(err, operation) {
            assert.strictEqual(operation.req.key, 10);
        },
        
        "Direction is 1": function(err, operation) {
            assert.strictEqual(operation.req.direction, 1);
        },
        
        "Count is 10": function(err, operation) {
            assert.strictEqual(operation.req.count, 10);
        }
    },
    "handleCollection(), read, invalid key": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "read",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                key: "Invalid key",
                count: 10,
                direction: 1
            });
        },
        
        "The store was not called": function(err, operation) {
            assert.ifError(operation.req);
        },
        
        "There is an error field": function(err, operation) {
            assert.ok(operation.data.error);
        }
    },
    
    "handleCollection(), read, invalid count": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "read",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                key: "+Infinity",
                count: "invalid",
                direction: -1
            });
        },
        
        "The store was not called": function(err, operation) {
            assert.ifError(operation.req);
        },
        
        "There is an error field": function(err, operation) {
            assert.ok(operation.data.error);
        }
    },
    
    "handleCollection(), read, missing direction": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "read",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                key: "-Infinity",
                count: 10
            });
        },
        
        "The store was not called": function(err, operation) {
            assert.ifError(operation.req);
        },
        
        "There is an error field": function(err, operation) {
            assert.ok(operation.data.error);
        }
    },
    "handleCollection(), read, invalid direction": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "read",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                key: "-Infinity",
                count: 10,
                direction: 0
            });
        },
        
        "The store was not called": function(err, operation) {
            assert.ifError(operation.req);
        },
        
        "There is an error field": function(err, operation) {
            assert.ok(operation.data.error);
        }
    },
    
    "handleCollection(), read, non-existing collection": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "read",
                collection: "Invalid",
                params: { foo: 1, bar: 1 },
                key: "-Infinity",
                count: 10,
                direction: 1
            });
        },
        
        "The store was not called": function(err, operation) {
            assert.ifError(operation.req);
        },
        
        "There is an error field": function(err, operation) {
            assert.ok(operation.data.error);
        }
    },
    
    "handleCollection(), read, invalid collection": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "read",
                collection: {},
                params: { foo: 1, bar: 1 },
                key: "-Infinity",
                count: 10,
                direction: 1
            });
        },
        
        "The store was not called": function(err, operation) {
            assert.ifError(operation.req);
        },
        
        "There is an error field": function(err, operation) {
            assert.ok(operation.data.error);
        }
    },
    "handleCollection(), read, invalid params": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "read",
                collection: "Test3",
                params: { foo: 1, invalid: 1 },
                key: "-Infinity",
                count: 10,
                direction: 1
            });
        },
        
        "The store was not called": function(err, operation) {
            assert.ifError(operation.req);
        },
        
        "There is an error field": function(err, operation) {
            assert.ok(operation.data.error);
        }
    },
    "handleCollection(), insert": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "insert",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                key: 1,
                data: { t: "Test" }
            });
        },
        
        "Insert operation was called": function(err, operation) {
            assert.equal(operation.req.operation, "insert");
        },
        
        "Key is 1": function(err, operation) {
            assert.strictEqual(operation.req.key, 1);
        },
        
        "Data is OK": function(err, operation) {
            assert.strictEqual(operation.req.data.t, "Test");
        }
    },
    
    "handleCollection(), insert, invalid key": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "insert",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                key: "-Infinity",
                data: { t: "Test" }
            });
        },
        
        "The store was not called": function(err, operation) {
            assert.ifError(operation.req);
        },
        
        "There is an error field": function(err, operation) {
            assert.ok(operation.data.error);
        }
    },
    "handleCollection(), insert, invalid data": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "insert",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                key: 10,
                data: "Invalid"
            });
        },
        
       "The store was not called": function(err, operation) {
           assert.ifError(operation.req);
       },
       
       "There is an error field": function(err, operation) {
           assert.ok(operation.data.error);
       }
    },
    "handleCollection(), delete": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "delete",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                id: 1
            });
        },
        
        "Delete operation was called": function(err, operation) {
            assert.equal(operation.req.operation, "delete");
        },
        
        "Id is 1": function(err, operation) {
            assert.strictEqual(operation.req.id, 1);
        }
    },
    "handleCollection(), delete, invalid id": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "delete",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                id: {}
            });
        },
        
        "The store was not called": function(err, operation) {
            assert.ifError(operation.req);
        },
        
        "There is an error field": function(err, operation) {
            assert.ok(operation.data.error);
        }
    },
    "handleCollection(), update": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "update",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                id: 1,
                key: 1,
                data: { t: "Test" }
            });
        },
        
        "Update operation was called": function(err, operation) {
            assert.equal(operation.req.operation, "update");
        },
        
        "Id is 1": function(err, operation) {
            assert.strictEqual(operation.req.id, 1);
        },
        "Key is 1": function(err, operation) {
            assert.strictEqual(operation.req.key, 1);
        },
        
        "Data is OK": function(err, operation) {
            assert.strictEqual(operation.req.data.t, "Test");
        }
    },
    "handleCollection(), update, invalid id": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "update",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                id: {},
                key: 1,
                data: { t: "Test" }
            });
        },
        
        "The store was not called": function(err, operation) {
            assert.ifError(operation.req);
        },
        
        "There is an error field": function(err, operation) {
            assert.ok(operation.data.error);
        }
    },
    "handleCollection(), update, invalid key": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "update",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                id: 1,
                key: "-Infinity",
                data: { t: "Test" }
            });
        },
        
        "The store was not called": function(err, operation) {
            assert.ifError(operation.req);
        },
        
        "There is an error field": function(err, operation) {
            assert.ok(operation.data.error);
        }
    },
    "handleCollection(), update, invalid data": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "update",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
                id: 1,
                key: 10,
                data: "Invalid"
            });
        },
        
       "The store was not called": function(err, operation) {
           assert.ifError(operation.req);
       },
       
       "There is an error field": function(err, operation) {
           assert.ok(operation.data.error);
       }
    },
    "handleCollection(), invalid message": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "invalid",
                collection: "Test3",
                params: { foo: 1, bar: 1 },
            });
        },
        
       "The store was not called": function(err, operation) {
           assert.ifError(operation.req);
       },
       
       "There is an error field": function(err, operation) {
           assert.ok(operation.data.error);
       }
    },

    "handleCollection(), subscribe": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "subscribe",
                collection: "Test3",
                params: { foo: 1, bar: 1 }
            });
        },
        
        "Socket joins the room": function(err, message) {
            assert.equal(message.join, 'collections/Test3/"foo":1,"bar":1');
        }
    },
    
    "handleCollection(), unsubscribe": {
        topic: function() {
            var manager = getManager(this.callback);
            var socket = new mock.Socket(this.callback);
            
            manager.handleCollection(socket, {
                operation: "unsubscribe",
                collection: "Test3",
                params: { foo: 1, bar: 1 }
            });
        },
        
        "Socket leaves the room": function(err, message) {
            assert.equal(message.leave, 'collections/Test3/"foo":1,"bar":1');
        }
    },
    
    "import, returns": {
        topic: function() {
            var manager = new Manager(new mock.Io);
            manager.add("Test", new mock.Store({}, this.callback), {
                import: function(data) {
                    return {
                        t2: data.t
                    };
                }
            });
            
            var socket = new mock.Socket(this.callback);

            manager.handleCollection(socket, {
                operation: "insert",
                collection: "Test",
                key: 1,
                data: { t: "Test" }
            });
        },
        
        "Insert operation was called": function(err, operation) {
            assert.equal(operation.req.operation, "insert");
        },
        
        "Data is OK": function(err, operation) {
            assert.strictEqual(operation.req.data.t2, "Test");
        }
    },
    
    "import, callback": {
        topic: function() {
            var manager = new Manager(new mock.Io);
            manager.add("Test", new mock.Store({}, this.callback), {
                import: function(data, callback) {
                    process.nextTick(function () {
                        callback(null, {
                            t2: data.t
                        });
                    });
                }
            });
            
            var socket = new mock.Socket(this.callback);

            manager.handleCollection(socket, {
                operation: "insert",
                collection: "Test",
                key: 1,
                data: { t: "Test" }
            });
        },
        
        "Insert operation was called": function(err, operation) {
            assert.equal(operation.req.operation, "insert");
        },
        
        "Data is OK": function(err, operation) {
            assert.strictEqual(operation.req.data.t2, "Test");
        }
    },
    
    "export, returns": {
        topic: function() {
            var manager = new Manager(new mock.Io());
            manager.add("Test", new collection.Store(), {
                export: function(data) {
                    return {
                        t2: data.t
                    };
                }
            });
            
            var socket = new mock.Socket(this.callback);

            manager.handleCollection(socket, {
                operation: "insert",
                collection: "Test",
                key: 1,
                data: { t: "Test" }
            });
        },
        
        "Insert operation was called": function(err, operation) {
            assert.equal(operation.data.operation, "insert");
        },
        
        "Data is OK": function(err, operation) {
            assert.strictEqual(operation.data.data.t2, "Test");
        }
    },
    
    "export, callback": {
        topic: function() {
            var manager = new Manager(new mock.Io);
            manager.add("Test", new collection.Store(), {
                export: function(data, callback) {
                    process.nextTick(function () {
                        callback(null, {
                            t2: data.t
                        });
                    });
                }
            });
            
            var socket = new mock.Socket(this.callback);

            manager.handleCollection(socket, {
                operation: "insert",
                collection: "Test",
                key: 1,
                data: { t: "Test" }
            });
        },
        
        "Insert operation was called": function(err, operation) {
            assert.equal(operation.data.operation, "insert");
        },
        
        "Data is OK": function(err, operation) {
            assert.strictEqual(operation.data.data.t2, "Test");
        }
    },
    
    "array export, returns": {
        topic: function() {
            var manager = new Manager(new mock.Io());
            manager.add("Test", {
                read: function(req, res) {
                    res(null, [
                        {id:0,key:0,data:{t:"Test0"}},
                        {id:1,key:1,data:{t:"Test1"}}
                    ]);
                }
            }, {
                export: function(data) {
                    return {
                        t2: data.t
                    };
                }
            });
            
            var socket = new mock.Socket(this.callback);

            manager.handleCollection(socket, {
                operation: "read",
                collection: "Test",
                key: 0,
                count: 2,
                direction: 1
            });
        },
        
        "Read operation was called": function(err, operation) {
            assert.equal(operation.data.operation, "read");
        },
        
        "Data is OK": function(err, operation) {
            assert.strictEqual(operation.data.data[0].data.t2, "Test0");
            assert.strictEqual(operation.data.data[1].data.t2, "Test1");
        }
    },
    
    "array export, callback": {
        topic: function() {
            var manager = new Manager(new mock.Io());
            manager.add("Test", {
                read: function(req, res) {
                    res(null, [
                        {id:0,key:0,data:{t:"Test0"}},
                        {id:1,key:1,data:{t:"Test1"}}
                    ]);
                }
            }, {
                export: function(data, callback) {
                    process.nextTick(function() { 
                        callback(null, {
                            t2: data.t
                        });
                    });
                }
            });
            
            var socket = new mock.Socket(this.callback);

            manager.handleCollection(socket, {
                operation: "read",
                collection: "Test",
                key: 0,
                count: 2,
                direction: 1
            });
        },
        
        "Read operation was called": function(err, operation) {
            assert.equal(operation.data.operation, "read");
        },
        
        "Data is OK": function(err, operation) {
            assert.strictEqual(operation.data.data[0].data.t2, "Test0");
            assert.strictEqual(operation.data.data[1].data.t2, "Test1");
        }
    },
              
    "validation success, returns": {
        topic: function() {
            var manager = new Manager(new mock.Io);
            manager.add("Test", new collection.Store(), {
                validate: function(id, key, data, callback) {
                    return { };
                }
            });
            
            var socket = new mock.Socket(this.callback);

            manager.handleCollection(socket, {
                operation: "insert",
                collection: "Test",
                key: 1,
                data: { t: "Test" }
            });
        },
        
        "Insert operation was called": function(err, operation) {
            assert.equal(operation.data.operation, "insert");
        },
        
        "Data is OK": function(err, operation) {
            assert.strictEqual(operation.data.data.t, "Test");
        }
    },
    
    "validation success, callback": {
        topic: function() {
            var manager = new Manager(new mock.Io);
            manager.add("Test", new collection.Store(), {
                validate: function(id, key, data, callback) {
                    process.nextTick(function () {
                        callback(null, { });
                    });
                }
            });
            
            var socket = new mock.Socket(this.callback);

            manager.handleCollection(socket, {
                operation: "insert",
                collection: "Test",
                key: 1,
                data: { t: "Test" }
            });
        },
        
        "Insert operation was called": function(err, operation) {
            assert.equal(operation.data.operation, "insert");
        },
        
        "Data is OK": function(err, operation) {
            assert.strictEqual(operation.data.data.t, "Test");
        }
    },
    
    "validation error, returns": {
        topic: function() {
            var manager = new Manager(new mock.Io);
            manager.add("Test", new collection.Store(), {
                validate: function(id, key, data, callback) {
                    return {
                        t: "Is wrong"
                    };
                }
            });
            
            var socket = new mock.Socket(this.callback);

            manager.handleCollection(socket, {
                operation: "insert",
                collection: "Test",
                key: 1,
                data: { t: "Test" }
            });
        },
        
        "Error operation was called": function(err, operation) {
            assert.equal(operation.data.operation, "error");
        },
        
        "Validation error is OK": function(err, operation) {
            assert.strictEqual(operation.data.fields.t, "Is wrong");
        }
    },
    
    "validation error, callback": {
        topic: function() {
            var manager = new Manager(new mock.Io);
            manager.add("Test", new collection.Store(), {
                validate: function(id, key, data, callback) {
                    process.nextTick(function () {
                        callback(null, {
                            t: "Is wrong"
                        });
                    });
                }
            });
            
            var socket = new mock.Socket(this.callback);

            manager.handleCollection(socket, {
                operation: "insert",
                collection: "Test",
                key: 1,
                data: { t: "Test" }
            });
        },
        
        "Error operation was called": function(err, operation) {
            assert.equal(operation.data.operation, "error");
        },
        
        "Validation error is OK": function(err, operation) {
            assert.strictEqual(operation.data.fields.t, "Is wrong");
        }
    },
    
    "throwing validator": {
        topic: function() {
            var manager = new Manager(new mock.Io);
            manager.add("Test", new collection.Store(), {
                validate: function(id, key, data, callback) {
                    throw new Error("Boo!");
                }
            });
            
            var socket = new mock.Socket(this.callback);

            manager.handleCollection(socket, {
                operation: "insert",
                collection: "Test",
                key: 1,
                data: { t: "Test" }
            });
        },
        
        "Error operation was called": function(err, operation) {
            assert.equal(operation.data.operation, "error");
        },
        
        "This is an internal error": function(err, operation) {
            assert.strictEqual(operation.data.error, "internal");
        }
    }
}).export(module);