/*! 
 * Collection
 *
 * Copyright (C) 2012 Philippe Plantier <philippe@midsummerweb.com>
 *
 * MIT Licensed
 */

var basedir;

if (process.env.JSCOVERAGE) {
    basedir = "../../var/jscoverage";
} else {
    basedir = "../../lib";
}

var vows = require("vows"),
    mock = require("./lib/mock"),
    assert = require("assert"),
    Memory = require(basedir + "/stores/memory"),
    Redis = require(basedir + "/stores/redis"),
    MongoDb = require(basedir + "/stores/mongodb"),
    async = require("async");
    
function getBatch(storeName, storeGenerator) {
    var batch = {};
    var testCounter = 0;
    
    batch[storeName + " - Insert one object"] = {
        topic: function() {
            var self = this;
            var store = storeGenerator("test" + testCounter++);
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({ key: 0, data: { t: "Test0" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.read({ key: -Infinity, count: 10000, direction: 1 }, new mock.Res(callback));
                }
            ], function(err, descriptors) {
                self.callback(err, descriptors);
            });
                
        },
        
        "Called send with an insert descriptor": function(err, descriptors) {
            assert.equal(typeof descriptors[1], "object");
            assert.equal(typeof descriptors[2], "object");
        },
        
        "The data actually was inserted in the array": function(err, descriptors) {        
            assert.equal(descriptors[descriptors.length-1].length, 1);
            assert.equal(descriptors[descriptors.length-1][0].key, 0);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test0");
        }
    };
    
    batch[storeName + " - Insert 2 objects"] = {
        topic: function() {
            var self = this;
            var store = storeGenerator("test" + testCounter++);
            
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({ key: 1, data: { t: "Test1" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 0, data: { t: "Test0" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.read({ key: -Infinity, count: 10000, direction: 1 }, new mock.Res(callback));
                }
                
            ], function(err, descriptors) {
                self.callback(err, descriptors);
            });
                
        },
        
        "4 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 4);
        },
        
        "The data in the array is correct and ordered": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 2);
            assert.equal(descriptors[descriptors.length-1][0].key, 0);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test0");
            assert.equal(descriptors[descriptors.length-1][1].key, 1);
            assert.equal(descriptors[descriptors.length-1][1].data.t, "Test1");
        }
    };
    
    batch[storeName + " - Insert 2 objects, then invert positions"] = {
        topic: function() {
            var self = this;
            var store = storeGenerator("test" + testCounter++);
            
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({ key: 1, data: { t: "Test1" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 0, data: { t: "Test0" }}, new mock.Res(callback));
                },
            ], function(err, descriptors1) {
                async.series([
                    function(callback) {
                        store.update({ id: descriptors1[2].id, key: 2, data: { t: "Test0" }}, new mock.Res(callback));
                    },
                    function(callback) {
                        store.read({ key: -Infinity, count: 10000, direction: 1 }, new mock.Res(callback));
                    }
                ], function(err, descriptors2) {
                    self.callback(err, descriptors1.concat(descriptors2));
                });
            }); 
        },
        
        "5 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 5);
        },
        
        "The 4th descriptor updates the object to the last pos": function(err, descriptors) {
            assert.equal(descriptors[3].id, descriptors[2].id);
            assert.equal(descriptors[3].key, 2);
        },
        
        "There are still 2 items": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 2);
        },
        
        "The data in the array is correct and ordered": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1][0].key, 1);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test1");
            assert.equal(descriptors[descriptors.length-1][1].key, 2);
            assert.equal(descriptors[descriptors.length-1][1].data.t, "Test0");
        }
    };
    
    batch[storeName + " - Insert 3 objects, then delete the middle one"] = {
        topic: function() {
            var self = this;
            var store = storeGenerator("test" + testCounter++);
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({ key: 0, data: { t: "Test0" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 1, data: { t: "Test1" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 2, data: { t: "Test2" }}, new mock.Res(callback));
                }
            ], function(err, descriptors1) {
                async.series([
                    function(callback) {
                        store.delete({ id: descriptors1[2].id }, new mock.Res(callback));
                    },
                    function(callback) {
                        store.read({ key: -Infinity, count: 10000, direction: 1 }, new mock.Res(callback));
                    }
                ], function(err, descriptors2) {
                    self.callback(err, descriptors1.concat(descriptors2));
                });
            }); 
                
        },
        
        "6 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 6);
        },
        
        "There are 2 items": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 2);
        },
        
        "The data in the array is correct and ordered": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1][0].key, 0);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test0");
            assert.equal(descriptors[descriptors.length-1][1].key, 2);
            assert.equal(descriptors[descriptors.length-1][1].data.t, "Test2");
        }
    };
    
    batch[storeName + " - Insert 3 objects, then delete a non-existing one"] = {
        topic: function() {
            var self = this;
            var store = storeGenerator("test" + testCounter++);
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({ key: 0, data: { t: "Test0" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 1, data: { t: "Test1" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 2, data: { t: "Test2" }}, new mock.Res(callback));
                }
            ], function(err, descriptors1) {
                async.series([
                    function(callback) {
                        store.delete({ id: descriptors1[2].id }, new mock.Res(callback));
                    },
                    function(callback) {
                        store.delete({ id: descriptors1[2].id }, new mock.Res(callback));
                    },
                    function(callback) {
                        store.read({ key: -Infinity, count: 10000, direction: 1 }, new mock.Res(callback));
                    }
                ], function(err, descriptors2) {
                    self.callback(err, descriptors1.concat(descriptors2));
                });
            }); 
                
        },
        
        "7 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 7);
        },
        
        "There are 2 items": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 2);
        },
        
        "The non-existing delete triggered an error": function(err, descriptors) {
            assert.ok(descriptors[5].error);
        },
        
        "The data in the array is correct and ordered": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1][0].key, 0);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test0");
            assert.equal(descriptors[descriptors.length-1][1].key, 2);
            assert.equal(descriptors[descriptors.length-1][1].data.t, "Test2");
        }
    };
    
    batch[storeName + " - Insert 3 objects, then update a non-existing one"] = {
        topic: function() {
            var self = this;
            var store = storeGenerator("test" + testCounter++);
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({ key: 0, data: { t: "Test0" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 1, data: { t: "Test1" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 2, data: { t: "Test2" }}, new mock.Res(callback));
                }
            ], function(err, descriptors1) {
                async.series([
                    function(callback) {
                        store.delete({ id: descriptors1[2].id }, new mock.Res(callback));
                    },
                    function(callback) {
                        store.update({ id: descriptors1[2].id, key: 3, data: { t: "Wrong" } }, new mock.Res(callback));
                    },
                    function(callback) {
                        store.read({ key: -Infinity, count: 10000, direction: 1 }, new mock.Res(callback));
                    }
                ], function(err, descriptors2) {
                    self.callback(err, descriptors1.concat(descriptors2));
                });
            }); 
                
        },
        
        "7 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 7);
        },
        
        "There are 2 items": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 2);
        },
        
        "The invalid update triggered an error": function(err, descriptors) {
            assert.ok(descriptors[5].error);
        },
        
        "The data in the array is correct and ordered": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1][0].key, 0);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test0");
            assert.equal(descriptors[descriptors.length-1][1].key, 2);
            assert.equal(descriptors[descriptors.length-1][1].data.t, "Test2");
        }
    };
    
    batch[storeName + " - Insert 3 objects, then delete an invalid ID"] = {
        topic: function() {
            var self = this;
            var store = storeGenerator("test" + testCounter++);
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({ key: 0, data: { t: "Test0" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 1, data: { t: "Test1" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 2, data: { t: "Test2" }}, new mock.Res(callback));
                }
            ], function(err, descriptors1) {
                async.series([
                    function(callback) {
                        store.delete({ id: "an invalid ID" }, new mock.Res(callback));
                    },
                    function(callback) {
                        store.read({ key: -Infinity, count: 10000, direction: 1 }, new mock.Res(callback));
                    }
                ], function(err, descriptors2) {
                    self.callback(err, descriptors1.concat(descriptors2));
                });
            }); 
                
        },
        
        "6 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 6);
        },
        
        "There are 3 items": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 3);
        },
        
        "The invalid delete triggered an error": function(err, descriptors) {
            assert.ok(descriptors[4].error);
        },
        
        "The data in the array is correct and ordered": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1][0].key, 0);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test0");
            assert.equal(descriptors[descriptors.length-1][1].key, 1);
            assert.equal(descriptors[descriptors.length-1][1].data.t, "Test1");
            assert.equal(descriptors[descriptors.length-1][2].key, 2);
            assert.equal(descriptors[descriptors.length-1][2].data.t, "Test2");
        }
    };
    
    batch[storeName + " - Insert 3 objects, then update an invalid ID"] = {
        topic: function() {
            var self = this;
            var store = storeGenerator("test" + testCounter++);
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({ key: 0, data: { t: "Test0" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 1, data: { t: "Test1" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 2, data: { t: "Test2" }}, new mock.Res(callback));
                }
            ], function(err, descriptors1) {
                async.series([
                    function(callback) {
                        store.update({ id: "Invalid ID", key: 3, data: { t: "Wrong" } }, new mock.Res(callback));
                    },
                    function(callback) {
                        store.read({ key: -Infinity, count: 10000, direction: 1 }, new mock.Res(callback));
                    }
                ], function(err, descriptors2) {
                    self.callback(err, descriptors1.concat(descriptors2));
                });
            }); 
                
        },
        
        "6 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 6);
        },
        
        "There are 3 items": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 3);
        },
        
        "The invalid update triggered an error": function(err, descriptors) {
            assert.ok(descriptors[4].error);
        },
        
        "The data in the array is correct and ordered": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1][0].key, 0);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test0");
            assert.equal(descriptors[descriptors.length-1][1].key, 1);
            assert.equal(descriptors[descriptors.length-1][1].data.t, "Test1");
            assert.equal(descriptors[descriptors.length-1][2].key, 2);
            assert.equal(descriptors[descriptors.length-1][2].data.t, "Test2");
        }
    };
    
    batch[storeName + " - Insert 3 objects, then concurrently delete and update"] = {
        topic: function() {
            var self = this;
            var store = storeGenerator("test" + testCounter++);
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({ key: 0, data: { t: "Test0" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 1, data: { t: "Test1" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 2, data: { t: "Test2" }}, new mock.Res(callback));
                }
            ], function(err, descriptors1) {
                async.series([
                    function(callback) {
                        async.parallel([
                            function(callback) {
                                store.delete({ id: descriptors1[2].id }, new mock.Res(callback));
                            },
                            function(callback) {
                                store.update({ id: descriptors1[2].id, key: 3, data: { t: "Test1" } }, new mock.Res(callback));
                            }
                        ], function(err, descriptors3) {
                            callback(err, descriptors3);
                        });
                    },
                    function(callback) {
                        store.read({ key: -Infinity, count: 10000, direction: 1 }, new mock.Res(callback));
                    }
                ], function(err, descriptors2) {
                    self.callback(err, descriptors1.concat(descriptors2));
                });
            }); 
                
        },
        
        "6 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 6);
        },
        
        "There are 2 items": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 2);
        },
        
        "The invalid update triggered an error": function(err, descriptors) {
            assert.equal(descriptors[4].length, 2);
            assert(descriptors[4][1].error);
        },
        
        "The data in the array is correct and ordered": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1][0].key, 0);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test0");
            assert.equal(descriptors[descriptors.length-1][1].key, 2);
            assert.equal(descriptors[descriptors.length-1][1].data.t, "Test2");
        }
    };
    
    batch[storeName + " - Insert 3 objects, then concurrently update and delete"] = {
        topic: function() {
            var self = this;
            var store = storeGenerator("test" + testCounter++);
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({ key: 0, data: { t: "Test0" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 1, data: { t: "Test1" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 2, data: { t: "Test2" }}, new mock.Res(callback));
                }
            ], function(err, descriptors1) {
                async.series([
                    function(callback) {
                        async.parallel([
                            function(callback) {
                                store.update({ id: descriptors1[2].id, key: 3, data: { t: "Test1" } }, new mock.Res(callback));
                            },
                            function(callback) {
                                store.delete({ id: descriptors1[2].id }, new mock.Res(callback));
                            }
                        ], function(err, descriptors3) {
                            callback(err, descriptors3);
                        });
                    },
                    function(callback) {
                        store.read({ key: -Infinity, count: 10000, direction: 1 }, new mock.Res(callback));
                    }
                ], function(err, descriptors2) {
                    self.callback(err, descriptors1.concat(descriptors2));
                });
            }); 
                
        },
        
        "6 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 6);
        },
        
        "There are 2 items": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 2);
        },
        
        "The delete operation was successful": function(err, descriptors) {
            assert.equal(descriptors[4].length, 2);
            assert.ok(!(descriptors[4][1] && descriptors[4][1].error));
        },
        
        "The data in the array is correct and ordered": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1][0].key, 0);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test0");
            assert.equal(descriptors[descriptors.length-1][1].key, 2);
            assert.equal(descriptors[descriptors.length-1][1].data.t, "Test2");
        }
    };
    
    batch[storeName + " - Insert 3 objects, then delete the middle one, sent as string"] = {
        topic: function() {
            var self = this;
            var store = storeGenerator("test" + testCounter++);
            
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({ key: 0, data: { t: "Test0" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 1, data: { t: "Test1" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 2, data: { t: "Test2" }}, new mock.Res(callback));
                } 
            ], function(err, descriptors1) {
                async.series([
                    function(callback) {
                        store.delete({ id: "" + descriptors1[2].id }, new mock.Res(callback));
                    },
                    function(callback) {
                        store.read({ key: -Infinity, count: 10000, direction: 1 }, new mock.Res(callback));
                    }  
                ], function(err, descriptors2) {
                    self.callback(err, descriptors1.concat(descriptors2));
                });
            }); 
            
                
        },
        
        "6 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 6);
        },
        
        "There are 2 items": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 2);
        },
        
        "The data in the array is correct and ordered": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1][0].key, 0);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test0");
            assert.equal(descriptors[descriptors.length-1][1].key, 2);
            assert.equal(descriptors[descriptors.length-1][1].data.t, "Test2");
        }
    };
    
    batch[storeName + " - Insert 3 objects, read from the middle"] = {
        topic: function() {
            var self = this;
            var store = storeGenerator("test" + testCounter++);
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({ key: 0, data: { t: "Test0" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 1, data: { t: "Test1" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 2, data: { t: "Test2" }}, new mock.Res(callback));
                }
            ], function(err, descriptors1) {
                async.series([
                    function(callback) {
                        store.read({ key: 1, count: 10000, direction: 1 }, new mock.Res(callback));
                    }
                ], function(err, descriptors2) {
                    self.callback(err, descriptors1.concat(descriptors2));
                });
            }); 
                
        },
        
        "5 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 5);
        },
        
        "There are 2 items": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 2);
        },
        
        "The data in the array is correct and ordered": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1][0].key, 1);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test1");
            assert.equal(descriptors[descriptors.length-1][1].key, 2);
            assert.equal(descriptors[descriptors.length-1][1].data.t, "Test2");
        }
    };
    
    batch[storeName + " - Insert 3 objects, inverted read from the middle"] = {
        topic: function() {
            var self = this;
            var store = storeGenerator("test" + testCounter++);
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({ key: 0, data: { t: "Test0" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 1, data: { t: "Test1" }}, new mock.Res(callback));
                },
                function(callback) {
                    store.insert({ key: 2, data: { t: "Test2" }}, new mock.Res(callback));
                }
            ], function(err, descriptors1) {
                async.series([
                    function(callback) {
                        store.read({ key: 1, count: 10000, direction: -1 }, new mock.Res(callback));
                    }
                ], function(err, descriptors2) {
                    self.callback(err, descriptors1.concat(descriptors2));
                });
            }); 
                
        },
        
        "5 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 5);
        },
        
        "There are 2 items": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 2);
        },
        
        "The data in the array is correct and ordered": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1][0].key, 0);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test0");
            assert.equal(descriptors[descriptors.length-1][1].key, 1);
            assert.equal(descriptors[descriptors.length-1][1].data.t, "Test1");
        }
    };
    
    batch[storeName + " - Use of params"] = {
        topic: function() {
            var self = this;
            var store = new storeGenerator("test" + testCounter++, { params: ["foo", "bar" ]});
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({
                        key: 0,
                        params: { 
                            foo: 1, 
                            bar: 2 
                        }, 
                        data: { t: "Test0" }
                    }, new mock.Res(callback));
                },
                function(callback) {
                    store.read({ 
                        key: -Infinity, 
                        count: 10000, 
                        direction: 1,
                        params: { 
                            foo: 1, 
                            bar: 2 
                        },
                    }, new mock.Res(callback));
                }
                
            ], function(err, descriptors) {
                self.callback(err, descriptors);
            });
                
        },
        
        "3 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 3);
        },
        
        "Called send with an insert descriptor": function(err, descriptors) {
            assert.equal(typeof descriptors[1], "object");
        },
        
        "The data actually was inserted in the array": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 1);
            assert.equal(descriptors[descriptors.length-1][0].key, 0);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test0");
        }
    };
    
    batch[storeName + " - Use of params (unsafe characters in params)"] = {
        topic: function() {
            var self = this;
            var store = new storeGenerator("test" + testCounter++, { params: ["foo", "bar" ]});
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({
                        key: 0,
                        params: { 
                            foo: 1, 
                            bar: "éà:/\\=*$+%{\x00}( )" 
                        }, 
                        data: { t: "Test0" }
                    }, new mock.Res(callback));
                },
                function(callback) {
                    store.read({ 
                        key: -Infinity, 
                        count: 10000, 
                        direction: 1,
                        params: { 
                            foo: 1, 
                            bar: "éà:/\\=*$+%{\x00}( )"  
                        },
                    }, new mock.Res(callback));
                }
                
            ], function(err, descriptors) {
                self.callback(err, descriptors);
            });
                
        },
        
        "3 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 3);
        },
        
        "Called send with an insert descriptor": function(err, descriptors) {
            assert.equal(typeof descriptors[1], "object");
        },
        
        "The data actually was inserted in the array": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 1);
            assert.equal(descriptors[descriptors.length-1][0].key, 0);
            assert.equal(descriptors[descriptors.length-1][0].data.t, "Test0");
        }
    };
    
    batch[storeName + " - Use of params (read with different params)"] = {
        topic: function() {
            var self = this;
            var store = new storeGenerator("test" + testCounter++, { params: ["foo", "bar" ]});
            async.series([
                function(callback) {
                    store.clear(callback);
                },
                function(callback) {
                    store.insert({
                        key: 0,
                        params: { 
                            foo: 1, 
                            bar: 2 
                        }, 
                        data: { t: "Test0" }
                    }, new mock.Res(callback));
                },
                function(callback) {
                    store.read({ 
                        key: -Infinity, 
                        count: 10000, 
                        direction: 1,
                        params: { 
                            foo: 1, 
                            bar: 3 
                        },
                    }, new mock.Res(callback));
                }
                
            ], function(err, descriptors) {
                self.callback(err, descriptors);
            });
                
        },
        
        "3 descriptors": function(err, descriptors) {
            assert.equal(descriptors.length, 3);
        },
        
        "Called send with an insert descriptor": function(err, descriptors) {
            assert.equal(typeof descriptors[1], "object");
        },
        
        "Thre is no data in the array read": function(err, descriptors) {
            assert.equal(descriptors[descriptors.length-1].length, 0);
        }
    };
    
    return batch;
}

var memoryBatch = getBatch("Memory", function (dbkey, options) {
    var settings = {};
    for (var key in options) {
        settings[key] = options[key];
    }
    
    return new Memory(settings);
});

var mongoBatch = getBatch("Mongo", function (dbkey, options) {
    var settings = {};
    for (var key in options) {
        settings[key] = options[key];
    }
    settings.db = "test";
    settings.collection = dbkey;
    return new MongoDb(settings);
});

var redisBatch = getBatch("Redis", function (dbkey, options) {
    var settings = {};
    for (var key in options) {
        settings[key] = options[key];
    }
    settings.key = dbkey;
    return new Redis(settings);
});

vows.describe("Common Stores Tests")
.addBatch(memoryBatch)
.addBatch(mongoBatch)
.addBatch(redisBatch)
.export(module);